from os import environ, path
from time import sleep

import pytest

from qbittorrentapi import APIConnectionError, Client, TorrentDictionary
from qbittorrentapi._version_support import (
    APP_VERSION_2_API_VERSION_MAP as api_version_map,
)

# Amount of time to attempt a check
CHECK_TIME = 10
# Amount of time to sleep between checks
CHECK_SLEEP = 0.25


def setup_environ():
    """Set up environment for testing; ensure qBittorrent is running."""
    environ.setdefault("QBITTORRENTAPI_HOST", "localhost:8080")
    environ.setdefault("QBITTORRENTAPI_USERNAME", "admin")
    environ.setdefault("QBITTORRENTAPI_PASSWORD", "adminadmin")
    try:
        environ.setdefault("QBT_VER", Client().app.version)
    except APIConnectionError:
        raise Exception("is qBittorrent running???")

    qbt_version = environ.get("QBT_VER", "")
    qbt_version = qbt_version if qbt_version.startswith("v") else f"v{qbt_version}"

    environ.setdefault("IS_QBT_DEV", "" if qbt_version in api_version_map else "1")
    is_qbt_dev = environ.get("IS_QBT_DEV", "false").lower() not in ["", "false"]

    return qbt_version, is_qbt_dev


def get_func(obj, method_name):
    """
    Retrieve a method from an object.

    For example, ``torrents_info`` or ``torrents.info``.
    """
    for attr in method_name.split("."):
        obj = getattr(obj, attr)
    return obj


def mkpath(*user_path):
    """Create the fully qualified path to an iterable of directories and/or file."""
    if any(user_path):
        return path.abspath(
            path.realpath(path.expanduser(path.join(*map(str, user_path))))
        )
    return ""


def retry(retries=3):
    """Decorator to retry a function if there's an exception."""

    def inner(f):
        def wrapper(*args, **kwargs):
            for retry_count in range(retries):
                try:
                    return f(*args, **kwargs)
                except Exception:
                    if retry_count >= (retries - 1):
                        raise

        return wrapper

    return inner


def get_torrent(client, torrent_hash) -> TorrentDictionary:
    """Retrieve a torrent from qBittorrent."""
    try:
        # not all versions of torrents_info() support passing a hash
        return [t for t in client.torrents_info() if t.hash == torrent_hash][0]
    except Exception:
        pytest.exit(f"Failed to find torrent for {torrent_hash}")


@retry(200)
def add_torrent(client, torrent_url, torrent_hash):
    client.torrents_add(urls=torrent_url, upload_limit=10, download_limit=10)
    if torrent_hash not in [t.hash for t in client.torrents_info()]:
        sleep(0.1)
        raise Exception("didn't find added torrent")


def check(check_func, value, reverse=False, negate=False, any=False, check_time=None):
    """
    Compare the return value of an arbitrary function to expected value with retries.
    Since some requests take some time to take effect in qBittorrent, the retries every
    second for 10 seconds.

    :param check_func: callable to generate values to check
    :param value: str, int, or iterator of values to look for
    :param reverse: False: look for check_func return in value; True: look for value in
        check_func return
    :param negate: False: value must be found; True: value must not be found
    :param check_time: maximum number of seconds to spend checking
    :param any: False: all values must be (not) found; True: any value must be (not)
        found
    """

    def _do_check(_check_func_val, _v, _negate, _reverse):
        if _negate:
            if _reverse:
                # print("Looking that %s is _not_ in %s" % (_v, _check_func_val))
                assert _v not in _check_func_val
            else:
                # print("Looking that %s is _not_ in %s" % (_check_func_val, (_v,)))
                assert _check_func_val not in (_v,)
        else:
            if _reverse:
                # print("Looking for %s in %s" % (_v, _check_func_val))
                assert _v in _check_func_val
            else:
                # print("Looking for %s in %s" % (_check_func_val, (_v,)))
                assert _check_func_val in (_v,)

    if isinstance(value, (str, int)):
        value = (value,)

    check_limit = int((check_time or CHECK_TIME) / CHECK_SLEEP)
    success = False

    try:
        for i in range(check_limit):
            try:
                exp = None
                for val in value:
                    # clear any previous exceptions if any=True
                    exp = None if any else exp

                    try:
                        # get val here so pytest includes value in failures
                        check_val = check_func()
                        _do_check(check_val, val, negate, reverse)
                    except AssertionError as e:
                        exp = e

                    # fail the test on first failure if any=False
                    if not any and exp:
                        break
                    # this value passed so test succeeded if any=True
                    if any and not exp:
                        break

                # raise caught inner exception for handling
                if exp:
                    raise exp

                # test succeeded!!!!
                success = True
                break

            except AssertionError:
                if i >= check_limit:
                    raise
                sleep(CHECK_SLEEP)
    except APIConnectionError:
        raise AssertionError("qBittorrent crashed...")

    if not success:
        raise Exception("Test neither succeeded nor failed...")



import glob
import os
from contextlib import contextmanager, suppress
from functools import partial
from os import environ, path
from sys import path as sys_path
from time import sleep
from unittest.mock import MagicMock

import pytest

from qbittorrentapi import APIConnectionError, Client
from qbittorrentapi._version_support import (
    APP_VERSION_2_API_VERSION_MAP as api_version_map,
)
from qbittorrentapi._version_support import v
from tests.utils import (
    CHECK_SLEEP,
    add_torrent,
    check,
    get_func,
    get_torrent,
    mkpath,
    retry,
    setup_environ,
)


class staticmethod:
    """Override staticmethod since it only become callable in 3.10."""

    def __init__(self, _func):
        self.f = _func

    def __call__(self, *a, **k):
        return self.f(*a, **k)


QBT_VERSION, IS_QBT_DEV = setup_environ()

BASE_PATH = sys_path[0]
RESOURCES_PATH = path.join(BASE_PATH, "tests", "_resources")
assert BASE_PATH.split("/")[-1] == "qbittorrent-api"

# fmt: off
ORIG_TORRENT_FILENAME = "ubuntu-22.04.1-desktop-amd64.iso.torrent"
ORIG_TORRENT_URL = f"https://github.com/rmartin16/qbittorrent-api/raw/main/tests/_resources/{ORIG_TORRENT_FILENAME}"
ORIG_TORRENT_HASH = "3b245504cf5f11bbdbe1201cea6a6bf45aee1bc0"
ORIG_TORRENT = None

TORRENT1_FILENAME = "kubuntu-22.04.4-desktop-amd64.iso.torrent"
TORRENT1_URL = f"https://github.com/rmartin16/qbittorrent-api/raw/main/tests/_resources/{TORRENT1_FILENAME}"
TORRENT1_HASH = "27a92b32757893ac9eb898e32c952636a3cc7b24"
TORRENT1_FILE_HANDLE = open(path.join(RESOURCES_PATH, TORRENT1_FILENAME), mode="rb")  # noqa: SIM115
TORRENT1_FILE = TORRENT1_FILE_HANDLE.read()

TORRENT2_FILENAME = "xubuntu-22.04.4-desktop-amd64.iso.torrent"
TORRENT2_URL = f"https://github.com/rmartin16/qbittorrent-api/raw/main/tests/_resources/{TORRENT2_FILENAME}"
TORRENT2_HASH = "c7d77fc3ecb68344b59ada11a0508dd6d08f2dfd"

ROOT_FOLDER_TORRENT_FILENAME = "root_folder.torrent"
ROOT_FOLDER_TORRENT_HASH = "a14553bd936a6d496402082454a70ea7a9521adc"
ROOT_FOLDER_TORRENT_FILE_HANDLE = open(path.join(RESOURCES_PATH, ROOT_FOLDER_TORRENT_FILENAME), mode="rb")  # noqa: E501, SIM115
ROOT_FOLDER_TORRENT_FILE = ROOT_FOLDER_TORRENT_FILE_HANDLE.read()
# fmt: on


@pytest.fixture(autouse=True)
def abort_if_qbittorrent_crashes(client):
    """Abort tests if qbittorrent seemingly disappears during testing."""
    try:
        client.app_version()
    except APIConnectionError:
        pytest.exit("qBittorrent crashed :(")


@pytest.fixture(autouse=True)
def skip_if_not_implemented(request, api_version):
    """Skips test if `skipif_before_api_version` marker specifies min API version."""
    if request.node.get_closest_marker("skipif_before_api_version"):
        version = request.node.get_closest_marker("skipif_before_api_version").args[0]
        if v(api_version) < v(version):
            pytest.skip(f"testing {v(api_version)}; needs {version} or later")


@pytest.fixture(autouse=True)
def skip_if_implemented(request, api_version):
    """Skips test if `skipif_after_api_version` marker specifies max API version."""
    if request.node.get_closest_marker("skipif_after_api_version"):
        version = request.node.get_closest_marker("skipif_after_api_version").args[0]
        if v(api_version) >= v(version):
            pytest.skip(f"testing {v(api_version)}; needs before {version}")


@pytest.fixture(scope="session")
def client():
    """qBittorrent Client for testing session."""
    client = Client(
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERBOSE_RESPONSE_LOGGING=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    )
    client.auth_log_in()
    client.app.preferences = dict(
        # enable RSS fetching
        rss_processing_enabled=True,
        # prevent banning IPs
        web_ui_max_auth_fail_count=1000,
        web_ui_ban_duration=1,
    )
    client.func = staticmethod(partial(get_func, client))
    try:
        add_torrent(client, ORIG_TORRENT_URL, ORIG_TORRENT_HASH)
    except Exception:
        pytest.exit("failed to add orig_torrent during setup")
    return client


@pytest.fixture
def client_mock(client):
    """qBittorrent Client for testing with request mocks."""
    client._get = MagicMock(wraps=client._get)
    client._get_cast = MagicMock(wraps=client._get_cast)
    client._post = MagicMock(wraps=client._post)
    client._post_cast = MagicMock(wraps=client._post_cast)
    try:
        yield client
    finally:
        client._get = client._get
        client._get_cast = client._get_cast
        client._post = client._post
        client._post_cast = client._post_cast


@pytest.fixture
def orig_torrent(client):
    """Torrent to remain in qBittorrent for entirety of session."""
    global ORIG_TORRENT
    if ORIG_TORRENT is None:
        ORIG_TORRENT = get_torrent(client, torrent_hash=ORIG_TORRENT_HASH)
        ORIG_TORRENT.func = staticmethod(partial(get_func, ORIG_TORRENT))
    ORIG_TORRENT.sync_local()  # ensure torrent is up-to-date
    return ORIG_TORRENT


@contextmanager
def new_torrent_standalone(client, torrent_hash=TORRENT1_HASH, tmp_path=None, **kwargs):
    def add_test_torrent(torrent_hash_, **kw):
        check_limit = int(10 / CHECK_SLEEP)
        for attempt in range(check_limit):
            if kw:
                client.torrents.add(**kw)
            elif tmp_path:
                client.torrents.add(
                    torrent_files=TORRENT1_FILE,
                    save_path=mkpath(tmp_path, "test_download"),
                    category="test_category",
                    is_paused=True,
                    upload_limit=1024,
                    download_limit=2048,
                    is_sequential_download=True,
                    is_first_last_piece_priority=True,
                )
            else:
                raise Exception("invalid params")
            try:
                torrent = get_torrent(client, torrent_hash_)
            except Exception:
                if attempt >= check_limit - 1:
                    raise
                sleep(CHECK_SLEEP)
            else:
                torrent.func = staticmethod(partial(get_func, torrent))
                return torrent

    @retry()
    def delete_test_torrent(client_, torrent_hash_):
        client_.torrents_delete(torrent_hashes=torrent_hash_, delete_files=True)
        check(
            lambda: [t.hash for t in client_.torrents_info()],
            torrent_hash_,
            reverse=True,
            negate=True,
        )

    try:
        yield add_test_torrent(torrent_hash, **kwargs)
    finally:
        delete_test_torrent(client, torrent_hash)


@pytest.fixture
def new_torrent(client, tmp_path):
    """Torrent that is added on demand to qBittorrent and then removed."""
    with new_torrent_standalone(client, tmp_path=tmp_path) as torrent:
        yield torrent


@pytest.fixture
def app_version(client):
    """qBittorrent Version being used for testing."""
    if IS_QBT_DEV:
        return client.app.version
    return QBT_VERSION


@pytest.fixture
def api_version(client):
    """qBittorrent Web API Version being used for testing."""
    if IS_QBT_DEV:
        return client.app.web_api_version
    return api_version_map[QBT_VERSION]


def pytest_sessionfinish(session, exitstatus):
    for fh in [TORRENT1_FILE_HANDLE, ROOT_FOLDER_TORRENT_FILE_HANDLE]:
        with suppress(Exception):
            fh.close()
    if environ.get("CI") != "true":
        client = Client()
        with suppress(Exception):
            # remove all torrents
            for torrent in client.torrents_info():
                torrent.delete(delete_files=True)
        # delete coverage files if not in CI
        for file in glob.iglob(path.join(BASE_PATH, ".coverage*")):
            os.unlink(file)
        # delete downloaded files if not in CI
        for filename in [TORRENT1_FILENAME, TORRENT2_FILENAME]:
            with suppress(Exception):
                os.unlink(mkpath("~", filename))



import sys

import pytest

from qbittorrentapi.definitions import APINames
from qbittorrentapi.log import LogMainList, LogPeersList


@pytest.mark.skipif(sys.version_info < (3, 9), reason="removeprefix not in 3.8")
def test_methods(client):
    namespace = APINames.Log
    all_dotted_methods = set(dir(getattr(client, namespace)))

    for meth in [meth for meth in dir(client) if meth.startswith(f"{namespace}_")]:
        assert meth.removeprefix(f"{namespace}_") in all_dotted_methods


@pytest.mark.parametrize("main_func", ["log_main", "log.main"])
@pytest.mark.parametrize("last_known_id", (None, 0))
def test_log_main_id(client, main_func, last_known_id):
    log_main = client.func(main_func)(last_known_id=last_known_id)
    assert isinstance(log_main, LogMainList)

    last_id = log_main[-1].id if log_main else 0
    log_main = client.func(main_func)(last_known_id=last_id)
    assert isinstance(log_main, LogMainList)
    assert not log_main or log_main[-1].id != last_id


@pytest.mark.parametrize("main_func", ["log_main", "log.main"])
def test_log_main_large_id(client, main_func):
    assert client.func(main_func)(last_known_id=99999999) == []


@pytest.mark.parametrize("main_func", ["log_main", "log.main"])
def test_log_main_slice(client, main_func):
    assert isinstance(client.func(main_func)()[1:2], LogMainList)


def test_log_main_info(client_mock):
    assert isinstance(client_mock.log.main.info(), LogMainList)
    client_mock._get_cast.assert_called_with(
        _name=APINames.Log,
        _method="main",
        params={
            "info": None,
            "normal": None,
            "warning": None,
            "critical": None,
            "last_known_id": None,
        },
        response_class=LogMainList,
    )


def test_log_main_normal(client_mock):
    assert isinstance(client_mock.log.main.normal(), LogMainList)
    client_mock._get_cast.assert_called_with(
        _name=APINames.Log,
        _method="main",
        params={
            "info": False,
            "normal": None,
            "warning": None,
            "critical": None,
            "last_known_id": None,
        },
        response_class=LogMainList,
    )


def test_log_main_warning(client_mock):
    assert isinstance(client_mock.log.main.warning(), LogMainList)
    client_mock._get_cast.assert_called_with(
        _name=APINames.Log,
        _method="main",
        params={
            "info": False,
            "normal": False,
            "warning": None,
            "critical": None,
            "last_known_id": None,
        },
        response_class=LogMainList,
    )


def test_log_main_critical(client_mock):
    assert isinstance(client_mock.log.main.critical(), LogMainList)
    client_mock._get_cast.assert_called_with(
        _name=APINames.Log,
        _method="main",
        params={
            "info": False,
            "normal": False,
            "warning": False,
            "critical": None,
            "last_known_id": None,
        },
        response_class=LogMainList,
    )


@pytest.mark.parametrize("main_func", ["log_main", "log.main"])
@pytest.mark.parametrize("include_level", (True, False, None, 1, 0))
def test_log_main_levels(client_mock, main_func, include_level):
    client_mock.func(main_func)(
        normal=include_level,
        info=include_level,
        warning=include_level,
        critical=include_level,
    )

    actual_include = None if include_level is None else bool(include_level)
    client_mock._get_cast.assert_called_with(
        _name=APINames.Log,
        _method="main",
        params={
            "normal": actual_include,
            "info": actual_include,
            "warning": actual_include,
            "critical": actual_include,
            "last_known_id": None,
        },
        response_class=LogMainList,
    )


@pytest.mark.parametrize("peers_func", ["log_peers", "log.peers"])
@pytest.mark.parametrize("last_known_id", (None, 0))
def test_log_peers_id(client, peers_func, last_known_id):
    log_peers = client.func(peers_func)(last_known_id=last_known_id)
    assert isinstance(log_peers, LogPeersList)

    last_id = log_peers[-1].id if log_peers else 0
    log_peers = client.func(peers_func)(last_known_id=last_id)
    assert isinstance(log_peers, LogPeersList)
    assert not log_peers or log_peers[-1].id != last_id


@pytest.mark.parametrize("peers_func", ["log_peers", "log.peers"])
def test_log_peers(client, peers_func):
    assert client.func(peers_func)(last_known_id=99999999) == []
    assert client.func(peers_func)(last_known_id=99999999) == []


@pytest.mark.parametrize("peers_func", ["log_peers", "log.peers"])
def test_log_peers_slice(client, peers_func):
    assert isinstance(client.func(peers_func)()[1:2], LogPeersList)



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import cast

from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.definitions import (
    APIKwargsT,
    APINames,
    ClientCache,
    Dictionary,
    JsonValueT,
    List,
    ListEntry,
    ListInputT,
)


class SearchResultsDictionary(Dictionary[JsonValueT]):
    """
    Response for :meth:`~SearchAPIMixIn.search_results`

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-search-results>`_
    """  # noqa: E501


class SearchStatus(ListEntry):
    """Item in :class:`SearchStatusesList`"""


class SearchStatusesList(List[SearchStatus]):
    """
    Response for :meth:`~SearchAPIMixIn.search_status`

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-search-status>`_
    """  # noqa: E501

    def __init__(self, list_entries: ListInputT, client: SearchAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=SearchStatus, client=client)


class SearchCategory(ListEntry):
    """Item in :class:`SearchCategoriesList`"""


class SearchCategoriesList(List[SearchCategory]):
    """Response for :meth:`~SearchAPIMixIn.search_categories`"""

    def __init__(self, list_entries: ListInputT, client: SearchAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=SearchCategory, client=client)


class SearchPlugin(ListEntry):
    """Item in :class:`SearchPluginsList`"""


class SearchPluginsList(List[SearchPlugin]):
    """
    Response for :meth:`~SearchAPIMixIn.search_plugins`.

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-search-plugins>`_
    """

    def __init__(self, list_entries: ListInputT, client: SearchAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=SearchPlugin, client=client)


class SearchAPIMixIn(AppAPIMixIn):
    """
    Implementation for all ``Search`` API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> search_job = client.search_start(pattern="Ubuntu", plugins="all", category="all")
        >>> client.search_stop(search_id=search_job.id)
        >>> # or
        >>> search_job.stop()
    """  # noqa: E501

    @property
    def search(self) -> Search:
        """
        Allows for transparent interaction with ``Search`` endpoints.

        See Search class for usage.
        """
        if self._search is None:
            self._search = Search(client=self)
        return self._search

    def search_start(
        self,
        pattern: str | None = None,
        plugins: str | Iterable[str] | None = None,
        category: str | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchJobDictionary:
        """
        Start a search. Python must be installed. Host may limit number of concurrent
        searches.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :raises Conflict409Error:

        :param pattern: term to search for
        :param plugins: list of plugins to use for searching (supports 'all' and
            'enabled')
        :param category: categories to limit search; dependent on plugins. (supports
            'all')
        """
        data = {
            "pattern": pattern,
            "plugins": self._list2string(plugins, "|"),
            "category": category,
        }
        return self._post_cast(
            _name=APINames.Search,
            _method="start",
            data=data,
            response_class=SearchJobDictionary,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_stop(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Stop a running search.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :raises NotFound404Error:
        :param search_id: ID of search job to stop
        """
        data = {"id": search_id}
        self._post(
            _name=APINames.Search,
            _method="stop",
            data=data,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_status(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchStatusesList:
        """
        Retrieve status of one or all searches.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :raises NotFound404Error:

        :param search_id: ID of search to get status; leave empty for status of all jobs
        """
        params = {"id": search_id}
        return self._get_cast(
            _name=APINames.Search,
            _method="status",
            params=params,
            response_class=SearchStatusesList,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_results(
        self,
        search_id: str | int | None = None,
        limit: str | int | None = None,
        offset: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchResultsDictionary:
        """
        Retrieve the results for the search.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :raises NotFound404Error:
        :raises Conflict409Error:

        :param search_id: ID of search job
        :param limit: number of results to return
        :param offset: where to start returning results
        """
        data = {"id": search_id, "limit": limit, "offset": offset}
        return self._post_cast(
            _name=APINames.Search,
            _method="results",
            data=data,
            response_class=SearchResultsDictionary,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_delete(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Delete a search job.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :raises NotFound404Error:
        :param search_id: ID of search to delete
        """
        data = {"id": search_id}
        self._post(
            _name=APINames.Search,
            _method="delete",
            data=data,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_categories(
        self,
        plugin_name: str | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchCategoriesList:
        """
        Retrieve categories for search.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1) and removed
        with qBittorrent v4.3.0 (Web API v2.6).

        :param plugin_name: Limit categories returned by plugin(s)
            (supports ``all`` and ``enabled``)
        """
        data = {"pluginName": plugin_name}
        return self._post_cast(
            _name=APINames.Search,
            _method="categories",
            data=data,
            response_class=SearchCategoriesList,
            version_introduced="2.1.1",
            version_removed="2.6",
            **kwargs,
        )

    def search_plugins(self, **kwargs: APIKwargsT) -> SearchPluginsList:
        """
        Retrieve details of search plugins.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).
        """
        return self._get_cast(
            _name=APINames.Search,
            _method="plugins",
            response_class=SearchPluginsList,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_install_plugin(
        self,
        sources: str | Iterable[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Install search plugins from either URL or file.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :param sources: list of URLs or filepaths
        """
        data = {"sources": self._list2string(sources, "|")}
        self._post(
            _name=APINames.Search,
            _method="installPlugin",
            data=data,
            version_introduced="2.1.1",
            **kwargs,
        )

    search_installPlugin = search_install_plugin

    def search_uninstall_plugin(
        self,
        names: str | Iterable[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Uninstall search plugins.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :param names: names of plugins to uninstall
        """
        data = {"names": self._list2string(names, "|")}
        self._post(
            _name=APINames.Search,
            _method="uninstallPlugin",
            data=data,
            version_introduced="2.1.1",
            **kwargs,
        )

    search_uninstallPlugin = search_uninstall_plugin

    def search_enable_plugin(
        self,
        plugins: str | Iterable[str] | None = None,
        enable: bool | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Enable or disable search plugin(s).

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :param plugins: list of plugin names
        :param enable: Defaults to ``True`` if ``None`` or unset;
            use ``False`` to disable
        """
        data = {
            "names": self._list2string(plugins, "|"),
            "enable": True if enable is None else bool(enable),
        }
        self._post(
            _name=APINames.Search,
            _method="enablePlugin",
            data=data,
            version_introduced="2.1.1",
            **kwargs,
        )

    search_enablePlugin = search_enable_plugin

    def search_update_plugins(self, **kwargs: APIKwargsT) -> None:
        """
        Auto update search plugins.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).
        """
        self._post(
            _name=APINames.Search,
            _method="updatePlugins",
            version_introduced="2.1.1",
            **kwargs,
        )

    search_updatePlugins = search_update_plugins

    def search_download_torrent(
        self,
        url: str | None = None,
        plugin: str | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Download a .torrent file or magnet for a search plugin.

        This method was introduced with qBittorrent v5.0.0 (Web API v2.11).

        :param url: URL for .torrent file or magnet
        :param plugin: Name of the plugin
        """
        data = {
            "torrentUrl": url,
            "pluginName": plugin,
        }
        self._post(
            _name=APINames.Search,
            _method="downloadTorrent",
            data=data,
            version_introduced="2.11",
            **kwargs,
        )

    search_downloadTorrent = search_download_torrent


class SearchJobDictionary(ClientCache[SearchAPIMixIn], Dictionary[JsonValueT]):
    """Response for :meth:`~SearchAPIMixIn.search_start`"""

    def __init__(self, data: Mapping[str, JsonValueT], client: SearchAPIMixIn):
        self._search_job_id: int | None = cast(int, data.get("id", None))
        super().__init__(data=data, client=client)

    def stop(self, **kwargs: APIKwargsT) -> None:
        """Implements :meth:`~SearchAPIMixIn.search_stop`."""
        self._client.search_stop(search_id=self._search_job_id, **kwargs)

    def status(self, **kwargs: APIKwargsT) -> SearchStatusesList:
        """Implements :meth:`~SearchAPIMixIn.search_status`."""
        return self._client.search_status(search_id=self._search_job_id, **kwargs)

    def results(
        self,
        limit: str | int | None = None,
        offset: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchResultsDictionary:
        """Implements :meth:`~SearchAPIMixIn.search_results`."""
        return self._client.search_results(
            limit=limit,
            offset=offset,
            search_id=self._search_job_id,
            **kwargs,
        )

    def delete(self, **kwargs: APIKwargsT) -> None:
        """Implements :meth:`~SearchAPIMixIn.search_delete`."""
        return self._client.search_delete(search_id=self._search_job_id, **kwargs)


class Search(ClientCache[SearchAPIMixIn]):
    """
    Allows interaction with ``Search`` API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> # this is all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'search_' prepended)
        >>> # initiate searches and retrieve results
        >>> search_job = client.search.start(pattern="Ubuntu", plugins="all", category="all")
        >>> status = search_job.status()
        >>> results = search_job.result()
        >>> search_job.delete()
        >>> # inspect and manage plugins
        >>> plugins = client.search.plugins
        >>> cats = client.search.categories(plugin_name="...")
        >>> client.search.install_plugin(sources="...")
        >>> client.search.update_plugins()
    """  # noqa: E501

    def start(
        self,
        pattern: str | None = None,
        plugins: str | Iterable[str] | None = None,
        category: str | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchJobDictionary:
        """Implements :meth:`~SearchAPIMixIn.search_start`."""
        return self._client.search_start(
            pattern=pattern,
            plugins=plugins,
            category=category,
            **kwargs,
        )

    def stop(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~SearchAPIMixIn.search_stop`."""
        return self._client.search_stop(search_id=search_id, **kwargs)

    def status(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchStatusesList:
        """Implements :meth:`~SearchAPIMixIn.search_status`."""
        return self._client.search_status(search_id=search_id, **kwargs)

    def results(
        self,
        search_id: str | int | None = None,
        limit: str | int | None = None,
        offset: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchResultsDictionary:
        """Implements :meth:`~SearchAPIMixIn.search_results`."""
        return self._client.search_results(
            search_id=search_id,
            limit=limit,
            offset=offset,
            **kwargs,
        )

    def delete(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~SearchAPIMixIn.search_delete`."""
        return self._client.search_delete(search_id=search_id, **kwargs)

    def categories(
        self,
        plugin_name: str | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchCategoriesList:
        """Implements :meth:`~SearchAPIMixIn.search_categories`."""
        return self._client.search_categories(plugin_name=plugin_name, **kwargs)

    @property
    def plugins(self) -> SearchPluginsList:
        """Implements :meth:`~SearchAPIMixIn.search_plugins`."""
        return self._client.search_plugins()

    def install_plugin(
        self,
        sources: str | Iterable[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~SearchAPIMixIn.search_install_plugin`."""
        return self._client.search_install_plugin(sources=sources, **kwargs)

    installPlugin = install_plugin

    def uninstall_plugin(
        self,
        sources: str | Iterable[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~SearchAPIMixIn.search_uninstall_plugin`."""
        return self._client.search_uninstall_plugin(sources=sources, **kwargs)

    uninstallPlugin = uninstall_plugin

    def enable_plugin(
        self,
        plugins: str | Iterable[str] | None = None,
        enable: bool | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~SearchAPIMixIn.search_enable_plugin`."""
        return self._client.search_enable_plugin(
            plugins=plugins,
            enable=enable,
            **kwargs,
        )

    enablePlugin = enable_plugin

    def update_plugins(self, **kwargs: APIKwargsT) -> None:
        """Implements :meth:`~SearchAPIMixIn.search_update_plugins`."""
        return self._client.search_update_plugins(**kwargs)

    updatePlugins = update_plugins

    def download_torrent(
        self,
        url: str | None = None,
        plugin: str | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~SearchAPIMixIn.search_download_torrent`."""
        return self._client.search_download_torrent(url=url, plugin=plugin)

    downloadTorrent = download_torrent



from qbittorrentapi._version_support import Version
from qbittorrentapi.app import (
    AppAPIMixIn,
    ApplicationPreferencesDictionary,
    BuildInfoDictionary,
    Cookie,
    CookieList,
    NetworkInterface,
    NetworkInterfaceAddressList,
    NetworkInterfaceList,
)
from qbittorrentapi.auth import AuthAPIMixIn
from qbittorrentapi.client import Client
from qbittorrentapi.definitions import (
    APINames,
    TorrentState,
    TorrentStates,
    TrackerStatus,
)
from qbittorrentapi.exceptions import (
    APIConnectionError,
    APIError,
    Conflict409Error,
    FileError,
    Forbidden403Error,
    HTTP4XXError,
    HTTP5XXError,
    HTTP400Error,
    HTTP401Error,
    HTTP403Error,
    HTTP404Error,
    HTTP409Error,
    HTTP415Error,
    HTTP500Error,
    HTTPError,
    InternalServerError500Error,
    InvalidRequest400Error,
    LoginFailed,
    MissingRequiredParameters400Error,
    NotFound404Error,
    TorrentFileError,
    TorrentFileNotFoundError,
    TorrentFilePermissionError,
    Unauthorized401Error,
    UnsupportedMediaType415Error,
    UnsupportedQbittorrentVersion,
)
from qbittorrentapi.log import LogAPIMixIn, LogEntry, LogMainList, LogPeer, LogPeersList
from qbittorrentapi.request import Request
from qbittorrentapi.rss import RSSAPIMixIn, RSSitemsDictionary, RSSRulesDictionary
from qbittorrentapi.search import (
    SearchAPIMixIn,
    SearchCategoriesList,
    SearchCategory,
    SearchJobDictionary,
    SearchPlugin,
    SearchPluginsList,
    SearchResultsDictionary,
    SearchStatus,
    SearchStatusesList,
)
from qbittorrentapi.sync import (
    SyncAPIMixIn,
    SyncMainDataDictionary,
    SyncTorrentPeersDictionary,
)
from qbittorrentapi.torrentcreator import (
    TaskStatus,
    TorrentCreator,
    TorrentCreatorAPIMixIn,
    TorrentCreatorTaskDictionary,
    TorrentCreatorTaskStatus,
    TorrentCreatorTaskStatusList,
)
from qbittorrentapi.torrents import (
    Tag,
    TagList,
    TorrentCategoriesDictionary,
    TorrentDictionary,
    TorrentFile,
    TorrentFilesList,
    TorrentInfoList,
    TorrentLimitsDictionary,
    TorrentPieceData,
    TorrentPieceInfoList,
    TorrentPropertiesDictionary,
    TorrentsAddPeersDictionary,
    TorrentsAPIMixIn,
    Tracker,
    TrackersList,
    WebSeed,
    WebSeedsList,
)
from qbittorrentapi.transfer import TransferAPIMixIn, TransferInfoDictionary

__all__ = (
    "APIConnectionError",
    "APIError",
    "APINames",
    "AppAPIMixIn",
    "ApplicationPreferencesDictionary",
    "AuthAPIMixIn",
    "BuildInfoDictionary",
    "Client",
    "Conflict409Error",
    "Cookie",
    "CookieList",
    "FileError",
    "Forbidden403Error",
    "HTTP400Error",
    "HTTP401Error",
    "HTTP403Error",
    "HTTP404Error",
    "HTTP409Error",
    "HTTP415Error",
    "HTTP4XXError",
    "HTTP500Error",
    "HTTP5XXError",
    "HTTPError",
    "InternalServerError500Error",
    "InvalidRequest400Error",
    "LogAPIMixIn",
    "LogEntry",
    "LoginFailed",
    "LogMainList",
    "LogPeer",
    "LogPeersList",
    "MissingRequiredParameters400Error",
    "NetworkInterface",
    "NetworkInterfaceList",
    "NetworkInterfaceAddressList",
    "NotFound404Error",
    "Request",
    "RSSAPIMixIn",
    "RSSitemsDictionary",
    "RSSRulesDictionary",
    "SearchAPIMixIn",
    "SearchCategoriesList",
    "SearchCategory",
    "SearchJobDictionary",
    "SearchPlugin",
    "SearchPluginsList",
    "SearchResultsDictionary",
    "SearchStatus",
    "SearchStatus",
    "SearchStatusesList",
    "SyncAPIMixIn",
    "SyncMainDataDictionary",
    "SyncTorrentPeersDictionary",
    "Tag",
    "TagList",
    "TaskStatus",
    "TorrentCreator",
    "TorrentCreatorAPIMixIn",
    "TorrentCreatorTaskDictionary",
    "TorrentCreatorTaskStatus",
    "TorrentCreatorTaskStatusList",
    "TorrentCategoriesDictionary",
    "TorrentDictionary",
    "TorrentFile",
    "TorrentFileError",
    "TorrentFileNotFoundError",
    "TorrentFilePermissionError",
    "TorrentFilesList",
    "TorrentInfoList",
    "TorrentLimitsDictionary",
    "TorrentPieceData",
    "TorrentPieceInfoList",
    "TorrentPropertiesDictionary",
    "TorrentsAddPeersDictionary",
    "TorrentsAPIMixIn",
    "TorrentState",
    "TorrentStates",
    "Tracker",
    "TrackersList",
    "TrackerStatus",
    "TransferAPIMixIn",
    "TransferInfoDictionary",
    "Unauthorized401Error",
    "UnsupportedMediaType415Error",
    "UnsupportedQbittorrentVersion",
    "Version",
    "WebSeed",
    "WebSeedsList",
)



import platform
from time import sleep
from types import MethodType
from unittest.mock import MagicMock

import pytest

from qbittorrentapi import (
    APINames,
    Conflict409Error,
    TorrentDictionary,
    TorrentFilesList,
    TorrentPieceInfoList,
    TorrentPropertiesDictionary,
    TorrentStates,
    TrackersList,
    WebSeedsList,
)
from qbittorrentapi._version_support import v
from tests.test_torrents import disable_queueing, enable_queueing
from tests.utils import check, mkpath, retry


def test_info(orig_torrent, monkeypatch):
    assert orig_torrent.info.hash == orig_torrent.hash
    # mimic <=v2.0.1 where torrents_info() doesn't support hash arg
    with monkeypatch.context() as m:
        m.setattr(orig_torrent._client, "app_version", MagicMock(return_value="2.0.0"))
        assert orig_torrent.info.hash == orig_torrent.hash

    # ensure if things are really broken, an empty TorrentDictionary is returned...
    if platform.python_implementation() == "CPython":
        with monkeypatch.context() as m:

            def fake_info(*args, **kwargs):
                return []

            m.setattr(orig_torrent._client, "torrents_info", fake_info)
            assert isinstance(orig_torrent.info, TorrentDictionary)
            assert bool(orig_torrent.info) is False


def test_sync_local(orig_torrent):
    orig_torrent.state = "gibberish"
    assert orig_torrent.state == "gibberish"
    orig_torrent.sync_local()
    assert isinstance(orig_torrent, TorrentDictionary)
    assert orig_torrent.state != "gibberish"


def test_state_enum(orig_torrent):
    assert orig_torrent.state_enum in TorrentStates
    assert orig_torrent.state_enum is not TorrentStates.UNKNOWN
    orig_torrent.resume()
    check(
        lambda: orig_torrent.sync_local() is None
        and orig_torrent.state_enum.is_downloading,
        True,
    )
    # simulate an unknown torrent.state
    orig_torrent.state = "gibberish"
    assert orig_torrent.state_enum is TorrentStates.UNKNOWN
    # restore torrent state
    orig_torrent.sync_local()
    check(lambda: orig_torrent.state_enum.is_downloading, True)


# test fails on 4.1.0 release
@pytest.mark.skipif_before_api_version("2.0.1")
def test_pause_resume(client, new_torrent):
    new_torrent.pause()
    check(
        lambda: client.torrents_info(torrent_hashes=new_torrent.hash)[
            0
        ].state_enum.is_paused,
        True,
    )

    new_torrent.resume()
    check(
        lambda: client.torrents_info(torrent_hashes=new_torrent.hash)[
            0
        ].state_enum.is_paused,
        False,
    )


# test fails on 4.1.0 release
@pytest.mark.skipif_before_api_version("2.0.1")
def test_stop_start(client, new_torrent):
    new_torrent.stop()
    check(
        lambda: client.torrents_info(torrent_hashes=new_torrent.hash)[
            0
        ].state_enum.is_paused,
        True,
    )

    new_torrent.resume()
    check(
        lambda: client.torrents_info(torrent_hashes=new_torrent.hash)[
            0
        ].state_enum.is_paused,
        False,
    )


@pytest.mark.parametrize("delete", [True, False, None, 0, 1])
def test_delete(client_mock, new_torrent, delete):
    new_torrent.delete(delete_files=delete)
    client_mock._post.assert_called_with(
        _name=APINames.Torrents,
        _method="delete",
        data={
            "hashes": new_torrent.hash,
            "deleteFiles": bool(delete),
        },
    )
    check(
        lambda: [t.hash for t in client_mock.torrents_info()],
        new_torrent.hash,
        reverse=True,
        negate=True,
    )


@pytest.mark.parametrize(
    "inc_prio_func, dec_prio_func, top_prio_func, bottom_prio_func",
    [
        ("increase_priority", "decrease_priority", "top_priority", "bottom_priority"),
        ("increasePrio", "decreasePrio", "topPrio", "bottomPrio"),
    ],
)
def test_priority(
    client, new_torrent, inc_prio_func, dec_prio_func, top_prio_func, bottom_prio_func
):
    disable_queueing(client)

    with pytest.raises(Conflict409Error):
        new_torrent.func(inc_prio_func)()
    with pytest.raises(Conflict409Error):
        new_torrent.func(dec_prio_func)()
    with pytest.raises(Conflict409Error):
        new_torrent.func(top_prio_func)()
    with pytest.raises(Conflict409Error):
        new_torrent.func(bottom_prio_func)()

    enable_queueing(client)
    sleep(0.25)  # putting sleeps in since these keep crashing qbittorrent

    current_priority = new_torrent.info.priority
    new_torrent.func(inc_prio_func)()
    sleep(0.25)
    check(lambda: new_torrent.info.priority < current_priority, True)

    current_priority = new_torrent.info.priority
    new_torrent.func(dec_prio_func)()
    sleep(0.25)
    check(lambda: new_torrent.info.priority > current_priority, True)

    current_priority = new_torrent.info.priority
    new_torrent.func(top_prio_func)()
    sleep(0.25)
    check(lambda: new_torrent.info.priority < current_priority, True)

    current_priority = new_torrent.info.priority
    new_torrent.func(bottom_prio_func)()
    sleep(0.25)
    check(lambda: new_torrent.info.priority > current_priority, True)


@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "set_share_limits_func", ["set_share_limits", "setShareLimits"]
)
def test_set_share_limits(orig_torrent, set_share_limits_func):
    orig_torrent.func(set_share_limits_func)(
        ratio_limit=5,
        seeding_time_limit=100,
        inactive_seeding_time_limit=200,
        share_limit_action="STOP",
    )
    check(lambda: orig_torrent.info.max_ratio, 5)
    check(lambda: orig_torrent.info.max_seeding_time, 100)
    if "max_inactive_seeding_time" in orig_torrent.info:
        check(lambda: orig_torrent.info.max_inactive_seeding_time, 200)


@pytest.mark.skipif_after_api_version("2.0.1")
@pytest.mark.parametrize(
    "set_share_limits_func", ["set_share_limits", "setShareLimits"]
)
def test_set_share_limits_not_implemented(orig_torrent, set_share_limits_func):
    with pytest.raises(NotImplementedError):
        orig_torrent.func(set_share_limits_func)(ratio_limit=5, seeding_time_limit=100)


@pytest.mark.parametrize(
    "down_limit_func, set_down_limit_func",
    [("download_limit", "set_download_limit"), ("downloadLimit", "setDownloadLimit")],
)
def test_download_limit(orig_torrent, down_limit_func, set_down_limit_func):
    setattr(orig_torrent, down_limit_func, 2048)
    check(lambda: orig_torrent.func(down_limit_func), 2048)
    check(lambda: orig_torrent.info.dl_limit, 2048)

    orig_torrent.func(set_down_limit_func)(4096)
    check(lambda: orig_torrent.func(down_limit_func), 4096)
    check(lambda: orig_torrent.info.dl_limit, 4096)


@pytest.mark.parametrize(
    "up_limit_func, set_up_limit_func",
    [("upload_limit", "set_upload_limit"), ("uploadLimit", "setUploadLimit")],
)
def test_upload_limit(orig_torrent, up_limit_func, set_up_limit_func):
    setattr(orig_torrent, up_limit_func, 2048)
    check(lambda: orig_torrent.func(up_limit_func), 2048)
    check(lambda: orig_torrent.info.up_limit, 2048)

    orig_torrent.func(set_up_limit_func)(4096)
    check(lambda: orig_torrent.func(up_limit_func), 4096)
    check(lambda: orig_torrent.info.up_limit, 4096)


@pytest.mark.skipif_before_api_version("2.0.2")
@pytest.mark.parametrize("set_loc_func", ["set_location", "setLocation"])
def test_set_location(new_torrent, set_loc_func, tmp_path):
    sleep(0.5)
    loc = mkpath(tmp_path, "3")
    new_torrent.func(set_loc_func)(loc)
    check(lambda: mkpath(new_torrent.info.save_path), mkpath(loc))


@pytest.mark.skipif_before_api_version("2.8.4")
@pytest.mark.parametrize("set_save_path_func", ["set_save_path", "setSavePath"])
def test_set_save_path(new_torrent, set_save_path_func, tmp_path):
    loc = mkpath(tmp_path, "savepath3")
    new_torrent.func(set_save_path_func)(loc)
    # qBittorrent may return trailing separators depending on version....
    check(lambda: mkpath(new_torrent.info.save_path), mkpath(loc))


@pytest.mark.skipif_before_api_version("2.8.4")
@pytest.mark.parametrize("set_down_path_func", ["set_download_path", "setDownloadPath"])
def test_set_download_path(new_torrent, set_down_path_func, tmp_path):
    loc = mkpath(tmp_path, "downloadpath3")
    new_torrent.func(set_down_path_func)(loc)
    # qBittorrent may return trailing separators depending on version....
    check(lambda: mkpath(new_torrent.info.download_path), mkpath(loc))


@pytest.mark.parametrize("set_cat_func", ["set_category", "setCategory"])
@pytest.mark.parametrize("category", ["category 1", "category_1"])
def test_set_category(client, orig_torrent, set_cat_func, category):
    client.torrents_create_category(category=category)
    orig_torrent.func(set_cat_func)(category=category)
    check(lambda: orig_torrent.info.category.replace("+", " "), category, reverse=True)
    client.torrents_remove_categories(categories=category)


@pytest.mark.parametrize(
    "set_auto_mgmt_func", ["set_auto_management", "setAutoManagement"]
)
def test_set_auto_management(orig_torrent, set_auto_mgmt_func):
    current_setting = orig_torrent.auto_tmm
    orig_torrent.func(set_auto_mgmt_func)(enable=(not current_setting))
    check(lambda: orig_torrent.info.auto_tmm, not current_setting)
    orig_torrent.func(set_auto_mgmt_func)(enable=current_setting)
    check(lambda: orig_torrent.info.auto_tmm, current_setting)


@pytest.mark.parametrize(
    "toggle_seq_down_func", ["toggle_sequential_download", "toggleSequentialDownload"]
)
def test_toggle_sequential_download(orig_torrent, toggle_seq_down_func):
    current_setting = orig_torrent.seq_dl
    orig_torrent.func(toggle_seq_down_func)()
    check(lambda: orig_torrent.info.seq_dl, not current_setting)
    orig_torrent.func(toggle_seq_down_func)()
    check(lambda: orig_torrent.info.seq_dl, current_setting)


@pytest.mark.skipif_before_api_version("2.0.2")
@pytest.mark.parametrize(
    "toggle_piece_prio_func",
    ["toggle_first_last_piece_priority", "toggleFirstLastPiecePrio"],
)
def test_toggle_first_last_piece_priority(orig_torrent, toggle_piece_prio_func):
    current_setting = orig_torrent.f_l_piece_prio
    orig_torrent.func(toggle_piece_prio_func)()
    check(lambda: orig_torrent.info.f_l_piece_prio, not current_setting)


@pytest.mark.parametrize("set_force_start_func", ["set_force_start", "setForceStart"])
def test_set_force_start(orig_torrent, set_force_start_func):
    current_setting = orig_torrent.force_start
    orig_torrent.func(set_force_start_func)(enable=(not current_setting))
    check(lambda: orig_torrent.info.force_start, not current_setting)
    orig_torrent.func(set_force_start_func)(enable=current_setting)
    check(lambda: orig_torrent.info.force_start, current_setting)


@pytest.mark.parametrize(
    "set_super_seeding_func", ["set_super_seeding", "setSuperSeeding"]
)
def test_set_super_seeding(orig_torrent, set_super_seeding_func):
    current_setting = orig_torrent.super_seeding
    orig_torrent.func(set_super_seeding_func)(enable=(not current_setting))
    check(lambda: orig_torrent.info.super_seeding, not current_setting)
    orig_torrent.func(set_super_seeding_func)(enable=current_setting)
    check(lambda: orig_torrent.info.super_seeding, current_setting)


def test_properties(orig_torrent):
    assert isinstance(orig_torrent.properties, TorrentPropertiesDictionary)
    assert "save_path" in orig_torrent.properties


@pytest.mark.parametrize("trackers", ["127.0.0.2", ["127.0.0.3", "127.0.0.4"]])
def test_trackers(orig_torrent, trackers):
    assert isinstance(orig_torrent.trackers, TrackersList)
    assert "num_peers" in orig_torrent.trackers[-1]

    orig_torrent.trackers = trackers
    check(lambda: (t.url for t in orig_torrent.trackers), trackers, reverse=True)


@pytest.mark.parametrize("add_trackers_func", ["add_trackers", "addTrackers"])
@pytest.mark.parametrize("trackers", ["127.0.0.2", ["127.0.0.3", "127.0.0.4"]])
def test_add_tracker(new_torrent, add_trackers_func, trackers):
    new_torrent.func(add_trackers_func)(urls=trackers)
    sleep(0.1)  # try to stop crashing qbittorrent
    check(lambda: (t.url for t in new_torrent.trackers), trackers, reverse=True)


@pytest.mark.skipif_before_api_version("2.2.0")
@pytest.mark.parametrize("edit_tracker_func", ["edit_tracker", "editTracker"])
def test_edit_tracker(orig_torrent, edit_tracker_func):
    orig_torrent.add_trackers(urls="127.0.1.1")
    orig_torrent.func(edit_tracker_func)(orig_url="127.0.1.1", new_url="127.0.1.2")
    check(
        lambda: (t.url for t in orig_torrent.trackers),
        "127.0.1.1",
        reverse=True,
        negate=True,
    )
    check(lambda: (t.url for t in orig_torrent.trackers), "127.0.1.2", reverse=True)
    orig_torrent.remove_trackers(urls="127.0.1.2")


@pytest.mark.skipif_after_api_version("2.2.0")
@pytest.mark.parametrize("edit_tracker_func", ["edit_tracker", "editTracker"])
def test_edit_tracker_not_implemented(orig_torrent, edit_tracker_func):
    with pytest.raises(NotImplementedError):
        orig_torrent.func(edit_tracker_func)()


@pytest.mark.skipif_before_api_version("2.2.0")
@pytest.mark.parametrize("remove_trackers_func", ["remove_trackers", "removeTrackers"])
@pytest.mark.parametrize("trackers", ["127.0.2.2", ["127.0.2.3", "127.0.2.4"]])
def test_remove_trackers(orig_torrent, remove_trackers_func, trackers):
    check(
        lambda: (t.url for t in orig_torrent.trackers),
        trackers,
        reverse=True,
        negate=True,
    )
    orig_torrent.add_trackers(urls=trackers)
    check(lambda: (t.url for t in orig_torrent.trackers), trackers, reverse=True)
    orig_torrent.func(remove_trackers_func)(urls=trackers)
    check(
        lambda: (t.url for t in orig_torrent.trackers),
        trackers,
        reverse=True,
        negate=True,
    )


@pytest.mark.skipif_after_api_version("2.2.0")
@pytest.mark.parametrize("remove_trackers_func", ["remove_trackers", "removeTrackers"])
def test_remove_trackers_not_implemented(orig_torrent, remove_trackers_func):
    with pytest.raises(NotImplementedError):
        orig_torrent.func(remove_trackers_func)()


def test_webseeds(orig_torrent):
    assert isinstance(orig_torrent.webseeds, WebSeedsList)


@pytest.mark.skipif_before_api_version("2.11.3")
@pytest.mark.parametrize("add_webseeds_func", ["add_webseeds", "addWebSeeds"])
@pytest.mark.parametrize(
    "webseeds",
    [
        "http://example/webseedone",
        ["http://example/webseedone", "http://example/webseedtwo"],
    ],
)
def test_add_webseed(new_torrent, add_webseeds_func, webseeds):
    assert new_torrent.webseeds == WebSeedsList([])
    new_torrent.func(add_webseeds_func)(urls=webseeds)
    assert sorted([w.url for w in new_torrent.webseeds]) == (
        webseeds if isinstance(webseeds, list) else [webseeds]
    )


@pytest.mark.skipif_before_api_version("2.11.3")
@pytest.mark.parametrize("edit_webseed_func", ["edit_webseed", "editWebSeed"])
def test_edit_webseeds(new_torrent, edit_webseed_func):
    assert new_torrent.webseeds == WebSeedsList([])
    new_torrent.add_webseeds(urls="http://example/asdf")
    new_torrent.edit_webseed(
        orig_url="http://example/asdf", new_url="http://example/vbnm"
    )
    assert new_torrent.webseeds[0].url == "http://example/vbnm"


@pytest.mark.skipif_before_api_version("2.11.3")
@pytest.mark.parametrize("remove_webseeds_func", ["remove_webseeds", "removeWebSeeds"])
@pytest.mark.parametrize(
    "webseeds",
    [
        "http://example/webseedone",
        ["http://example/webseedone", "http://example/webseedtwo"],
    ],
)
def test_remove_webseeds(client, new_torrent, remove_webseeds_func, webseeds):
    assert new_torrent.webseeds == WebSeedsList([])
    new_torrent.add_webseeds(
        urls=[
            "http://example/webseedone",
            "http://example/webseedtwo",
            "http://example/webseedthree",
        ]
    )
    new_torrent.func(remove_webseeds_func)(urls=webseeds)
    for webseed in webseeds if isinstance(webseeds, list) else [webseeds]:
        assert webseed not in {w.url for w in new_torrent.webseeds}


def test_files(orig_torrent):
    assert isinstance(orig_torrent.files, TorrentFilesList)
    assert "id" in orig_torrent.files[0]


def test_recheck(orig_torrent):
    orig_torrent.recheck()


@pytest.mark.skipif_before_api_version("2.0.2")
def test_reannounce(orig_torrent):
    orig_torrent.reannounce()


@pytest.mark.skipif_after_api_version("2.0.2")
def test_reannounce_not_implemented(orig_torrent):
    with pytest.raises(NotImplementedError):
        orig_torrent.reannounce()


@pytest.mark.skipif_before_api_version("2.10")
def test_reannounce_in(orig_torrent):
    assert "reannounce_in" in orig_torrent.info
    assert type(orig_torrent.reannounce) is MethodType


@pytest.mark.skipif_after_api_version("2.10")
def test_reannounce_in_not_present(orig_torrent):
    assert "reannounce_in" not in orig_torrent.info


@pytest.mark.skipif_before_api_version("2.4.0")
@pytest.mark.parametrize("rename_file_func", ["rename_file", "renameFile"])
@pytest.mark.parametrize("name", ["new_name", "new name"])
def test_rename_file(app_version, new_torrent, rename_file_func, name):
    @retry()
    def run_test_old():
        new_torrent.func(rename_file_func)(file_id=0, new_file_name=name)
        check(lambda: new_torrent.files[0].name, name)

    run_test_old()

    if v(app_version) >= v("v4.3.3"):

        @retry()
        def run_test_new():
            curr_name = new_torrent.files[0].name
            new_name = "NEW_" + name
            new_torrent.func(rename_file_func)(old_path=curr_name, new_path=new_name)
            check(lambda: new_torrent.files[0].name, new_name)

        run_test_new()


@pytest.mark.skipif_after_api_version("2.4.0")
@pytest.mark.parametrize("rename_file_func", ["rename_file", "renameFile"])
def test_rename_file_not_implemented(new_torrent, rename_file_func):
    with pytest.raises(NotImplementedError):
        new_torrent.func(rename_file_func)()


@pytest.mark.skipif_before_api_version("2.7")
@pytest.mark.parametrize("rename_folder_func", ["rename_folder", "renameFolder"])
@pytest.mark.parametrize("name", ["new_name", "new name"])
def test_rename_folder(app_version, new_torrent, rename_folder_func, name):
    @retry()
    def test():
        # need to ensure we're at least on v4.3.3 to run test since
        # both v4.3.2 and v4.3.2 both use Web API 2.7
        if v(app_version) >= v("v4.3.3"):
            # move the file in to a new folder
            orig_file_path = new_torrent.files[0].name
            new_folder = "qwer"
            new_torrent.rename_file(
                old_path=orig_file_path,
                new_path=new_folder + "/" + orig_file_path,
            )

            # wait for the folder to be renamed
            check(
                lambda: [f.name.split("/")[0] for f in new_torrent.files],
                new_folder,
                reverse=True,
            )

            # test rename that new folder
            new_torrent.func(rename_folder_func)(
                old_path=new_folder,
                new_path=name,
            )
            check(
                lambda: new_torrent.files[0].name.replace("+", " "),
                name + "/" + orig_file_path,
            )

    test()


@pytest.mark.skipif_after_api_version("2.4.0")
@pytest.mark.parametrize("rename_folder_func", ["rename_folder", "renameFolder"])
def test_rename_folder_not_implemented(new_torrent, rename_folder_func):
    with pytest.raises(NotImplementedError):
        new_torrent.func(rename_folder_func)()


@pytest.mark.skipif_before_api_version("2.8.14")
def test_export(orig_torrent):
    assert isinstance(orig_torrent.export(), bytes)


@pytest.mark.skipif_after_api_version("2.8.14")
def test_export_not_implemented(orig_torrent):
    with pytest.raises(NotImplementedError):
        orig_torrent.export()


@pytest.mark.parametrize("piece_states_func", ["piece_states", "pieceStates"])
def test_piece_states(orig_torrent, piece_states_func):
    assert isinstance(orig_torrent.func(piece_states_func), TorrentPieceInfoList)


@pytest.mark.parametrize("piece_hashes_func", ["piece_hashes", "pieceHashes"])
def test_piece_hashes(orig_torrent, piece_hashes_func):
    assert isinstance(orig_torrent.func(piece_hashes_func), TorrentPieceInfoList)


@pytest.mark.parametrize("file_prio_func", ["file_priority", "filePriority"])
def test_file_priority(orig_torrent, file_prio_func):
    orig_torrent.func(file_prio_func)(file_ids=0, priority=7)
    check(lambda: orig_torrent.files[0].priority, 7)


@pytest.mark.parametrize("name", ["new_name", "new name"])
def test_rename(new_torrent, name):
    new_torrent.rename(new_name=name)
    check(lambda: new_torrent.info.name.replace("+", " "), name)


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_tags_func, remove_tags_func",
    [("add_tags", "remove_tags"), ("addTags", "removeTags")],
)
@pytest.mark.parametrize("tags", ["tag 1", ["tag 2", "tag 3"]])
def test_add_remove_tags(client, orig_torrent, add_tags_func, remove_tags_func, tags):
    try:
        orig_torrent.func(add_tags_func)(tags=tags)
        check(lambda: orig_torrent.info.tags, tags, reverse=True)

        orig_torrent.func(remove_tags_func)(tags=tags)
        check(lambda: orig_torrent.info.tags, tags, reverse=True, negate=True)
    finally:
        client.torrents_delete_tags(tags=tags)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_tags_func, remove_tags_func",
    [("add_tags", "remove_tags"), ("addTags", "removeTags")],
)
def test_add_remove_tags_not_implemented(
    client, orig_torrent, add_tags_func, remove_tags_func
):
    with pytest.raises(NotImplementedError):
        orig_torrent.func(add_tags_func)()
    with pytest.raises(NotImplementedError):
        orig_torrent.func(remove_tags_func)()


@pytest.mark.skipif_before_api_version("2.11.4")
@pytest.mark.parametrize("set_tags_func", ["set_tags", "setTags"])
@pytest.mark.parametrize("tags", ["tag 1", ["tag 2", "tag 3"]])
def test_set_tags(client, orig_torrent, set_tags_func, tags):
    try:
        orig_torrent.add_tags(tags="extra-tag")
        orig_torrent.func(set_tags_func)(tags=tags)
        check(lambda: orig_torrent.info.tags, tags, reverse=True)
    finally:
        client.torrents_delete_tags(tags=tags)


@pytest.mark.skipif_after_api_version("2.11.4")
@pytest.mark.parametrize("set_tags_func", ["set_tags", "setTags"])
def test_set_tags_not_implemented(client, orig_torrent, set_tags_func):
    with pytest.raises(NotImplementedError):
        orig_torrent.func(set_tags_func)()



import sys
from pathlib import Path

import pytest

from qbittorrentapi import APINames
from qbittorrentapi.exceptions import NotFound404Error
from qbittorrentapi.torrentcreator import (
    TaskStatus,
    TorrentCreatorTaskDictionary,
    TorrentCreatorTaskStatus,
    TorrentCreatorTaskStatusList,
)
from tests.utils import check


@pytest.fixture
def task(client) -> TorrentCreatorTaskDictionary:
    source_path = Path(Path.cwd() / "tests/_resources/test_torrent")
    source_path.mkdir(parents=True, exist_ok=True)
    test_file = source_path / "test-file"
    test_file.touch()
    with test_file.open("w") as f:
        f.write("hello world")

    yield (
        task := client.torrentcreator_add_task(
            source_path=Path("/tmp/_resources/test_torrent"), start_seeding=False
        )
    )

    task.delete()
    test_file.unlink()
    source_path.rmdir()


@pytest.mark.skipif(sys.version_info < (3, 9), reason="removeprefix not in 3.8")
def test_methods(client):
    namespace = APINames.TorrentCreator
    all_dotted_methods = set(dir(getattr(client, namespace)))

    for meth in [meth for meth in dir(client) if meth.startswith(f"{namespace}_")]:
        assert meth.removeprefix(f"{namespace}_") in all_dotted_methods


@pytest.mark.skipif_before_api_version("2.10.4")
@pytest.mark.parametrize(
    "add_task_func",
    [
        "torrentcreator_add_task",
        "torrentcreator_addTask",
        "torrentcreator.add_task",
        "torrentcreator.addTask",
    ],
)
def test_add_task(client, add_task_func):
    task = client.func(add_task_func)(source_path="/empty-dir", start_seeding=False)

    assert isinstance(task, TorrentCreatorTaskDictionary)
    assert task.taskID
    assert task.task_id


@pytest.mark.skipif_after_api_version("2.10.4")
@pytest.mark.parametrize(
    "add_task_func",
    [
        "torrentcreator_add_task",
        "torrentcreator_addTask",
        "torrentcreator.add_task",
        "torrentcreator.addTask",
    ],
)
def test_add_task_not_implemented(client, add_task_func):
    with pytest.raises(NotImplementedError):
        client.func(add_task_func)()


@pytest.mark.skipif_before_api_version("2.10.4")
@pytest.mark.parametrize(
    "status_func", ["torrentcreator_status", "torrentcreator.status"]
)
def test_status(client, status_func, task):
    assert isinstance(
        client.func(status_func)(task_id=task.task_id), TorrentCreatorTaskStatusList
    )
    assert isinstance(
        client.func(status_func)(task_id=task.task_id)[0], TorrentCreatorTaskStatus
    )
    assert isinstance(task.status(), TorrentCreatorTaskStatus)
    assert TaskStatus(task.status().status) in TaskStatus


@pytest.mark.skipif_after_api_version("2.10.4")
@pytest.mark.parametrize(
    "status_func", ["torrentcreator_status", "torrentcreator.status"]
)
def test_status_not_implemented(client, status_func):
    with pytest.raises(NotImplementedError):
        client.func(status_func)()


def test_status_enum():
    assert TaskStatus("Failed") is TaskStatus.FAILED
    assert TaskStatus("Queued") is TaskStatus.QUEUED
    assert TaskStatus("Running") is TaskStatus.RUNNING
    assert TaskStatus("Finished") is TaskStatus.FINISHED


@pytest.mark.skipif_before_api_version("2.10.4")
@pytest.mark.parametrize(
    "torrent_file_func", ["torrentcreator_torrent_file", "torrentcreator.torrent_file"]
)
def test_torrent_file(client, torrent_file_func, task):
    check(lambda: TaskStatus(task.status().status), [TaskStatus.FINISHED])
    assert isinstance(client.func(torrent_file_func)(task_id=task.task_id), bytes)
    assert isinstance(task.torrent_file(), bytes)
    assert isinstance(task.torrentFile(), bytes)


@pytest.mark.skipif_after_api_version("2.10.4")
@pytest.mark.parametrize(
    "torrent_file_func", ["torrentcreator_torrent_file", "torrentcreator.torrent_file"]
)
def test_torrent_file_not_implemented(client, torrent_file_func):
    with pytest.raises(NotImplementedError):
        client.func(torrent_file_func)()


@pytest.mark.skipif_before_api_version("2.10.4")
@pytest.mark.parametrize(
    "torrent_delete_func", ["torrentcreator_delete_task", "torrentcreator.delete_task"]
)
def test_delete(client, torrent_delete_func):
    task = client.torrentcreator_add_task(source_path="/empty-dir", start_seeding=False)
    task.delete()
    with pytest.raises(NotFound404Error):
        task.status()

    task = client.torrentcreator_add_task(source_path="/empty-dir", start_seeding=False)
    client.func(torrent_delete_func)(task_id=task.task_id)
    with pytest.raises(NotFound404Error):
        client.func(torrent_delete_func)(task_id=task.task_id)


@pytest.mark.skipif_after_api_version("2.10.4")
@pytest.mark.parametrize(
    "torrent_delete_func", ["torrentcreator_delete_task", "torrentcreator.delete_task"]
)
def test_delete_not_implemented(client, torrent_delete_func):
    with pytest.raises(NotImplementedError):
        client.func(torrent_delete_func)()



from __future__ import annotations

from requests.exceptions import HTTPError as RequestsHTTPError
from requests.exceptions import RequestException


class APIError(Exception):
    """Base error for all exceptions from this Client."""


class UnsupportedQbittorrentVersion(APIError):
    """Connected qBittorrent is not fully supported by this Client."""


class FileError(IOError, APIError):
    """Base class for all exceptions for file handling."""


class TorrentFileError(FileError):
    """Base class for all exceptions for torrent files."""


class TorrentFileNotFoundError(TorrentFileError):
    """Specified torrent file does not appear to exist."""


class TorrentFilePermissionError(TorrentFileError):
    """Permission was denied to read the specified torrent file."""


class APIConnectionError(RequestException, APIError):
    """Base class for all communications errors including HTTP errors."""


class LoginFailed(APIConnectionError):
    """This can technically be raised with any request since log in may be attempted for
    any request and could fail."""


class HTTPError(RequestsHTTPError, APIConnectionError):
    """
    Base error for all HTTP errors.

    All errors following a successful connection to qBittorrent are returned as HTTP
    statuses.
    """

    http_status_code: int


class HTTP4XXError(HTTPError):
    """Base error for all HTTP 4XX statuses."""


class HTTP5XXError(HTTPError):
    """Base error for all HTTP 5XX statuses."""


class HTTP400Error(HTTP4XXError):
    """HTTP 400 Status."""

    http_status_code: int = 400


class HTTP401Error(HTTP4XXError):
    """HTTP 401 Status."""

    http_status_code: int = 401


class HTTP403Error(HTTP4XXError):
    """HTTP 403 Status."""

    http_status_code: int = 403


class HTTP404Error(HTTP4XXError):
    """HTTP 404 Status."""

    http_status_code: int = 404


class HTTP405Error(HTTP4XXError):
    """HTTP 405 Status."""

    http_status_code: int = 405


class HTTP409Error(HTTP4XXError):
    """HTTP 409 Status."""

    http_status_code: int = 409


class HTTP415Error(HTTP4XXError):
    """HTTP 415 Status."""

    http_status_code: int = 415


class HTTP500Error(HTTP5XXError):
    """HTTP 500 Status."""

    http_status_code: int = 500


class MissingRequiredParameters400Error(HTTP400Error):
    """Endpoint call is missing one or more required parameters."""


class InvalidRequest400Error(HTTP400Error):
    """One or more endpoint arguments are malformed."""


class Unauthorized401Error(HTTP401Error):
    """Primarily reserved for XSS and host header issues."""


class Forbidden403Error(HTTP403Error):
    """Not logged in, IP has been banned, or calling an API method that isn't public."""


class NotFound404Error(HTTP404Error):
    """This should mean qBittorrent couldn't find a torrent for the torrent hash."""


class MethodNotAllowed405Error(HTTP405Error):
    """HTTP method is not supported for the API endpoint."""


class Conflict409Error(HTTP409Error):
    """Returned if arguments don't make sense specific to the endpoint."""


class UnsupportedMediaType415Error(HTTP415Error):
    """``torrents/add`` endpoint will return this for invalid URL(s) or files."""


class InternalServerError500Error(HTTP500Error):
    """Returned if qBittorrent errors internally while processing the request."""



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from datetime import datetime
from importlib.metadata import version as metadata_version

base_path = os.path.abspath("../..")
sys.path.insert(0, os.path.join(base_path, "src"))


# -- Project information -----------------------------------------------------
project = "qbittorrent-api"
author = "Russell Martin"
copyright = f"{datetime.today().year}, {author}"

# The full version, including alpha/beta/rc tags
release = version = metadata_version("qbittorrent-api")

# -- General configuration ---------------------------------------------------
# Add any Sphinx extension module names here, as strings. They can be extensions coming
# with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.githubpages",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.intersphinx",
    "sphinx_copybutton",
]

pygments_style = "sphinx"

# warn about everything
nitpicky = True
nitpick_ignore = [
    ("py:class", "JsonValueT"),
    ("py:class", "ListInputT"),
    ("py:class", "Response"),
    ("py:class", "Request"),
    ("py:class", "qbittorrentapi.request.ResponseT"),
    ("py:class", "qbittorrentapi.request.T"),
    ("py:class", "qbittorrentapi.torrents.TorrentFilesT"),
    ("py:obj", "qbittorrentapi._attrdict.K"),
    ("py:obj", "qbittorrentapi._attrdict.V"),
    ("py:obj", "qbittorrentapi.definitions.K"),
    ("py:obj", "qbittorrentapi.definitions.V"),
]

autodoc_type_aliases = {"JsonValueT": "qbittorrentapi.definitions.JsonValueT"}
always_use_bars_union = True
add_module_names = False
autodoc_typehints_format = "short"
python_use_unqualified_type_names = True
python_use_unqualified_names = True
typehints_fully_qualified = False
typehints_use_signature = False
typehints_use_signature_return = True
typehints_document_rtype = True

source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

linkcheck_ignore = [
    # ignore reference to just the HTTP schemas
    r"^http://$",
    r"^https://$",
    r"^http://localhost",
    r"^http://example.com",
    # RTD CI seems to be banned...
    r"^https://distrowatch.com/news/torrents.xml",
]

# ignore GitHub code line number anchors that are generated by javascript
linkcheck_anchors_ignore = ["L79", "L42"]

# -- Options for HTML output -------------------------------------------------
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

html_theme = "furo"

# Add mappings
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "requests": ("https://requests.readthedocs.io/en/latest/", None),
}

suppress_warnings = [
    # Suppress builds warnings while building epub
    "epub.unknown_project_files",
]

# -- Options for spelling -------------------------------------------

# Spelling check needs an additional module that is not installed by default.
# Add it only if spelling check is requested so docs can be generated without it.
if "spelling" in sys.argv:
    extensions.append("sphinxcontrib.spelling")

# Spelling language.
spelling_lang = "en_US"

# Location of word list.
spelling_word_list_filename = "spelling_wordlist"

spelling_ignore_pypi_package_names = False



from __future__ import annotations

from collections import UserList
from collections.abc import Iterable, Mapping, Sequence
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    Union,
)

from qbittorrentapi._attrdict import AttrDict

if TYPE_CHECKING:
    from qbittorrentapi import Request

K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")

#: Type to define JSON.
JsonValueT = Union[
    None,
    int,
    str,
    bool,
    Sequence["JsonValueT"],
    Mapping[str, "JsonValueT"],
]
#: Type ``Any`` for ``kwargs`` parameters for API methods.
APIKwargsT = Any
#: Type for this API Client.
ClientT = TypeVar("ClientT", bound="Request")
#: Type for entry in List from API.
ListEntryT = TypeVar("ListEntryT", bound="ListEntry")
#: Type for List input to API method.
ListInputT = Iterable[Mapping[str, JsonValueT]]
#: Type for Files input to API method.
FilesToSendT = Mapping[str, Union[bytes, tuple[str, bytes]]]


class APINames(str, Enum):
    """
    API namespaces for API endpoints.

    e.g ``torrents`` in ``http://localhost:8080/api/v2/torrents/addTrackers``
    """

    Authorization = "auth"
    Application = "app"
    Log = "log"
    Sync = "sync"
    Transfer = "transfer"
    Torrents = "torrents"
    TorrentCreator = "torrentcreator"
    RSS = "rss"
    Search = "search"
    EMPTY = ""


class TorrentState(str, Enum):
    """
    Torrent States as defined by qBittorrent.

    Note: In qBittorrent v5.0.0:
     - ``PAUSED_UPLOAD`` was renamed to ``STOPPED_UPLOAD``
     - ``PAUSED_DOWNLOAD`` was renamed to ``STOPPED_DOWNLOAD``

    Definitions:
        - wiki: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-torrent-list>`_
        - code: `<https://github.com/qbittorrent/qBittorrent/blob/8e6515be2c8cc2b335002ab8913e9dcdd7873204/src/base/bittorrent/torrent.h#L79>`_

    :Usage:
        >>> from qbittorrentapi import Client, TorrentState
        >>> client = Client()
        >>> # print torrent hashes for torrents that are downloading
        >>> for torrent in client.torrents_info():
        >>>     # check if torrent is downloading
        >>>     if torrent.state_enum.is_downloading:
        >>>         print(f'{torrent.hash} is downloading...')
        >>>     # the appropriate enum member can be directly derived
        >>>     state_enum = TorrentState(torrent.state)
        >>>     print(f'{torrent.hash}: {state_enum.value}')
    """

    ERROR = "error"
    MISSING_FILES = "missingFiles"
    UPLOADING = "uploading"
    #: ``pausedUP`` was renamed to ``stoppedUP`` in Web API v2.11.0
    PAUSED_UPLOAD = "pausedUP"
    STOPPED_UPLOAD = "stoppedUP"
    QUEUED_UPLOAD = "queuedUP"
    STALLED_UPLOAD = "stalledUP"
    CHECKING_UPLOAD = "checkingUP"
    FORCED_UPLOAD = "forcedUP"
    ALLOCATING = "allocating"
    DOWNLOADING = "downloading"
    METADATA_DOWNLOAD = "metaDL"
    FORCED_METADATA_DOWNLOAD = "forcedMetaDL"
    #: ``pausedDL`` was renamed to ``stoppedDL`` in Web API v2.11.0
    PAUSED_DOWNLOAD = "pausedDL"
    STOPPED_DOWNLOAD = "stoppedDL"
    QUEUED_DOWNLOAD = "queuedDL"
    FORCED_DOWNLOAD = "forcedDL"
    STALLED_DOWNLOAD = "stalledDL"
    CHECKING_DOWNLOAD = "checkingDL"
    CHECKING_RESUME_DATA = "checkingResumeData"
    MOVING = "moving"
    UNKNOWN = "unknown"

    @property
    def is_downloading(self) -> bool:
        """Returns ``True`` if the State is categorized as Downloading."""
        return self in {
            TorrentState.DOWNLOADING,
            TorrentState.METADATA_DOWNLOAD,
            TorrentState.FORCED_METADATA_DOWNLOAD,
            TorrentState.STALLED_DOWNLOAD,
            TorrentState.CHECKING_DOWNLOAD,
            TorrentState.PAUSED_DOWNLOAD,
            TorrentState.STOPPED_DOWNLOAD,
            TorrentState.QUEUED_DOWNLOAD,
            TorrentState.FORCED_DOWNLOAD,
        }

    @property
    def is_uploading(self) -> bool:
        """Returns ``True`` if the State is categorized as Uploading."""
        return self in {
            TorrentState.UPLOADING,
            TorrentState.STALLED_UPLOAD,
            TorrentState.CHECKING_UPLOAD,
            TorrentState.QUEUED_UPLOAD,
            TorrentState.FORCED_UPLOAD,
        }

    @property
    def is_complete(self) -> bool:
        """Returns ``True`` if the State is categorized as Complete."""
        return self in {
            TorrentState.UPLOADING,
            TorrentState.STALLED_UPLOAD,
            TorrentState.CHECKING_UPLOAD,
            TorrentState.PAUSED_UPLOAD,
            TorrentState.STOPPED_UPLOAD,
            TorrentState.QUEUED_UPLOAD,
            TorrentState.FORCED_UPLOAD,
        }

    @property
    def is_checking(self) -> bool:
        """Returns ``True`` if the State is categorized as Checking."""
        return self in {
            TorrentState.CHECKING_UPLOAD,
            TorrentState.CHECKING_DOWNLOAD,
            TorrentState.CHECKING_RESUME_DATA,
        }

    @property
    def is_errored(self) -> bool:
        """Returns ``True`` if the State is categorized as Errored."""
        return self in {TorrentState.MISSING_FILES, TorrentState.ERROR}

    @property
    def is_stopped(self) -> bool:
        """Returns ``True`` if the State is categorized as Stopped."""
        return self in {
            TorrentState.PAUSED_UPLOAD,
            TorrentState.STOPPED_UPLOAD,
            TorrentState.PAUSED_DOWNLOAD,
            TorrentState.STOPPED_DOWNLOAD,
        }

    @property
    def is_paused(self) -> bool:
        """Alias of :any:`TorrentState.is_stopped`"""
        return self.is_stopped


TorrentStates = TorrentState


class TrackerStatus(int, Enum):
    """
    Tracker Statuses as defined by qBittorrent.

    Definitions:
        - wiki: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-torrent-trackers>`_
        - code: `<https://github.com/qbittorrent/qBittorrent/blob/5dcc14153f046209f1067299494a82e5294d883a/src/base/bittorrent/trackerentry.h#L42>`_

    :Usage:
        >>> from qbittorrentapi import Client, TrackerStatus
        >>> client = Client()
        >>> # print torrent hashes for torrents that are downloading
        >>> for torrent in client.torrents_info():
        >>>     for tracker in torrent.trackers:
        >>>         # display status for each tracker
        >>>         print(f"{torrent.hash[-6:]}: {TrackerStatus(tracker.status).display:>13} :{tracker.url}")
    """  # noqa: E501

    DISABLED = 0
    NOT_CONTACTED = 1
    WORKING = 2
    UPDATING = 3
    NOT_WORKING = 4

    @property
    def display(self) -> str:
        """Returns a descriptive display value for status."""
        return {
            TrackerStatus.DISABLED: "Disabled",
            TrackerStatus.NOT_CONTACTED: "Not contacted",
            TrackerStatus.WORKING: "Working",
            TrackerStatus.UPDATING: "Updating",
            TrackerStatus.NOT_WORKING: "Not working",
        }[self]


class ClientCache(Generic[ClientT]):
    """
    Caches the client.

    Subclass this for any object that needs access to the Client.
    """

    def __init__(self, *args: Any, client: ClientT, **kwargs: Any):
        self._client = client
        super().__init__(*args, **kwargs)


class Dictionary(AttrDict[V]):
    """Base definition of dictionary-like objects returned from qBittorrent."""

    def __init__(self, data: Mapping[str, JsonValueT] | None = None, **kwargs: Any):
        super().__init__(self._normalize(data or {}))
        # allows updating properties that aren't necessarily a part of the AttrDict
        self._setattr("_allow_invalid_attributes", True)

    @classmethod
    def _normalize(cls, data: Mapping[str, V] | T) -> AttrDict[V] | T:
        """Iterate through a dict converting any nested dicts to AttrDicts."""
        if isinstance(data, Mapping):
            return AttrDict({key: cls._normalize(value) for key, value in data.items()})
        return data


class List(UserList[ListEntryT]):
    """Base definition for list-like objects returned from qBittorrent."""

    def __init__(
        self,
        list_entries: ListInputT | None = None,
        entry_class: type[ListEntryT] | None = None,
        **kwargs: Any,
    ):
        super().__init__(
            [
                (
                    entry_class(data=entry, **kwargs)  # type: ignore[misc]
                    if entry_class is not None and isinstance(entry, Mapping)
                    else entry
                )
                for entry in list_entries or []
            ]
        )


class ListEntry(Dictionary[JsonValueT]):
    """Base definition for objects within a list returned from qBittorrent."""



from __future__ import annotations

import os
from collections.abc import Iterable, Mapping, Sequence
from json import dumps
from logging import Logger, getLogger
from typing import Any, AnyStr, Union

from qbittorrentapi.auth import AuthAPIMixIn
from qbittorrentapi.definitions import (
    APIKwargsT,
    APINames,
    ClientCache,
    Dictionary,
    JsonValueT,
    List,
    ListEntry,
    ListInputT,
)

logger: Logger = getLogger(__name__)


class ApplicationPreferencesDictionary(Dictionary[JsonValueT]):
    """
    Response for :meth:`~AppAPIMixIn.app_preferences`

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-application-preferences>`_
    """  # noqa: E501


class BuildInfoDictionary(Dictionary[Union[str, int]]):
    """
    Response for :meth:`~AppAPIMixIn.app_build_info`

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-build-info>`_
    """  # noqa: E501


class Cookie(ListEntry):
    """Item in :class:`CookieList`"""


class CookieList(List[Cookie]):
    """Response for :meth:`~AppAPIMixIn.app_cookies`"""

    def __init__(self, list_entries: ListInputT, client: AppAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=Cookie)


class NetworkInterface(ListEntry):
    """Item in :class:`NetworkInterfaceList`"""


class NetworkInterfaceList(List[NetworkInterface]):
    """Response for :meth:`~AppAPIMixIn.app_network_interface_list`"""

    def __init__(self, list_entries: ListInputT, client: AppAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=NetworkInterface)


# only API response that's a list of strings...so just ignore the typing for now
class NetworkInterfaceAddressList(List[str]):  # type: ignore
    """Response for :meth:`~AppAPIMixIn.app_network_interface_address_list`"""

    def __init__(self, list_entries: Iterable[str], client: AppAPIMixIn | None = None):
        super().__init__(list_entries)  # type: ignore


class DirectoryContentList(List[Union[str, ListEntry]]):  # type: ignore[type-var]
    """Response for :meth:`~AppAPIMixIn.app_get_directory_content`"""

    def __init__(self, list_entries: Iterable[str], client: AppAPIMixIn | None = None):
        super().__init__(list_entries)  # type: ignore


class AppAPIMixIn(AuthAPIMixIn):
    """
    Implementation of all ``Application`` API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> client.app_version()
        >>> client.app_preferences()
    """  # noqa: E501

    @property
    def app(self) -> Application:
        """
        Allows for transparent interaction with Application endpoints.

        See Application class for usage.
        """
        if self._application is None:
            self._application = Application(client=self)
        return self._application

    @property
    def application(self) -> Application:
        return self.app

    def app_version(self, **kwargs: APIKwargsT) -> str:
        """qBittorrent application version."""
        return self._get_cast(
            _name=APINames.Application,
            _method="version",
            response_class=str,
            **kwargs,
        )

    def app_web_api_version(self, **kwargs: APIKwargsT) -> str:
        """qBittorrent Web API version."""
        return self._get_cast(
            _name=APINames.Application,
            _method="webapiVersion",
            response_class=str,
            **kwargs,
        )

    app_webapiVersion = app_web_api_version

    def app_build_info(self, **kwargs: APIKwargsT) -> BuildInfoDictionary:
        """
        qBittorrent build info.

        This method was introduced with qBittorrent v4.2.0 (Web API v2.3).
        """
        return self._get_cast(
            _name=APINames.Application,
            _method="buildInfo",
            response_class=BuildInfoDictionary,
            version_introduced="2.3",
            **kwargs,
        )

    app_buildInfo = app_build_info

    def app_shutdown(self, **kwargs: APIKwargsT) -> None:
        """Shutdown qBittorrent."""
        self._post(_name=APINames.Application, _method="shutdown", **kwargs)

    def app_preferences(self, **kwargs: APIKwargsT) -> ApplicationPreferencesDictionary:
        """Retrieve qBittorrent application preferences."""
        return self._get_cast(
            _name=APINames.Application,
            _method="preferences",
            response_class=ApplicationPreferencesDictionary,
            **kwargs,
        )

    def app_set_preferences(
        self,
        prefs: ApplicationPreferencesDictionary | Mapping[str, Any] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Set one or more preferences in qBittorrent application.

        :param prefs: dictionary of preferences to set
        """
        data = {"json": dumps(prefs, separators=(",", ":"))}
        self._post(
            _name=APINames.Application,
            _method="setPreferences",
            data=data,
            **kwargs,
        )

    app_setPreferences = app_set_preferences

    def app_default_save_path(self, **kwargs: APIKwargsT) -> str:
        """The default path where torrents are saved."""
        return self._get_cast(
            _name=APINames.Application,
            _method="defaultSavePath",
            response_class=str,
            **kwargs,
        )

    app_defaultSavePath = app_default_save_path

    def app_cookies(self, **kwargs: APIKwargsT) -> CookieList:
        """Retrieve current cookies."""
        return self._get_cast(
            _name=APINames.Application,
            _method="cookies",
            response_class=CookieList,
            version_introduced="2.11.3",
            **kwargs,
        )

    def app_set_cookies(
        self,
        cookies: CookieList | Sequence[Mapping[str, str | int]] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Set cookies.

        .. highlight:: python
        .. code-block:: python

            cookies = [
                {
                    'domain': 'example.com',
                    'path': '/example/path',
                    'name': 'cookie name',
                    'value': 'cookie value',
                    'expirationDate': 1729366667,
                },
            ]

        :param cookies: list of cookies to set
        """
        data = {
            # coerce to types that are known to be JSON serializable
            "cookies": dumps(list(map(dict, cookies or ())), separators=(",", ":")),
        }
        self._post(
            _name=APINames.Application,
            _method="setCookies",
            data=data,
            version_introduced="2.11.3",
            **kwargs,
        )

    app_setCookies = app_set_cookies

    def app_network_interface_list(self, **kwargs: APIKwargsT) -> NetworkInterfaceList:
        """
        The list of network interfaces on the host.

        This method was introduced with qBittorrent v4.2.0 (Web API v2.3).
        """
        return self._get_cast(
            _name=APINames.Application,
            _method="networkInterfaceList",
            response_class=NetworkInterfaceList,
            version_introduced="2.3",
            **kwargs,
        )

    app_networkInterfaceList = app_network_interface_list

    def app_network_interface_address_list(
        self,
        interface_name: str = "",
        **kwargs: APIKwargsT,
    ) -> NetworkInterfaceAddressList:
        """
        The addresses for a network interface; omit name for all addresses.

        This method was introduced with qBittorrent v4.2.0 (Web API v2.3).

        :param interface_name: Name of interface to retrieve addresses for
        """
        data = {"iface": interface_name}
        return self._post_cast(
            _name=APINames.Application,
            _method="networkInterfaceAddressList",
            data=data,
            response_class=NetworkInterfaceAddressList,
            version_introduced="2.3",
            **kwargs,
        )

    app_networkInterfaceAddressList = app_network_interface_address_list

    def app_send_test_email(self) -> None:
        """Sends a test email using the configured email address."""
        self._post(
            _name=APINames.Application,
            _method="sendTestEmail",
            version_introduced="2.10.4",
        )

    app_sendTestEmail = app_send_test_email

    def app_get_directory_content(
        self,
        directory_path: str | os.PathLike[AnyStr] | None = None,
        with_metadata: bool | None = None,
    ) -> DirectoryContentList:
        """
        The contents of a directory file path.

        :raises NotFound404Error: file path not found or not a directory
        :param directory_path: file system path to directory
        :param with_metadata: include file/directory metadata
        """
        data = {
            "dirPath": (
                os.fsdecode(directory_path) if directory_path is not None else None
            ),
            "withMetadata": with_metadata,
        }
        return self._post_cast(
            _name=APINames.Application,
            _method="getDirectoryContent",
            data=data,
            response_class=DirectoryContentList,
            version_introduced="2.11",
        )

    app_getDirectoryContent = app_get_directory_content


class Application(ClientCache[AppAPIMixIn]):
    """
    Allows interaction with ``Application`` API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> # these are all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'app_' prepended)
        >>> webapiVersion = client.application.webapiVersion
        >>> web_api_version = client.application.web_api_version
        >>> app_web_api_version = client.application.web_api_version
        >>> # access and set preferences as attributes
        >>> is_dht_enabled = client.application.preferences.dht
        >>> # supports sending a just subset of preferences to update
        >>> client.application.preferences = dict(dht=(not is_dht_enabled))
        >>> prefs = client.application.preferences
        >>> prefs["web_ui_clickjacking_protection_enabled"] = True
        >>> client.app.preferences = prefs
        >>>
        >>> client.application.shutdown()
    """  # noqa: E501

    @property
    def version(self) -> str:
        """Implements :meth:`~AppAPIMixIn.app_version`."""
        return self._client.app_version()

    @property
    def web_api_version(self) -> str:
        """Implements :meth:`~AppAPIMixIn.app_web_api_version`."""
        return self._client.app_web_api_version()

    webapiVersion = web_api_version

    @property
    def build_info(self) -> BuildInfoDictionary:
        """Implements :meth:`~AppAPIMixIn.app_build_info`."""
        return self._client.app_build_info()

    buildInfo = build_info

    def shutdown(self, **kwargs: APIKwargsT) -> None:
        """Implements :meth:`~AppAPIMixIn.app_shutdown`."""
        self._client.app_shutdown(**kwargs)

    @property
    def preferences(self) -> ApplicationPreferencesDictionary:
        """Implements :meth:`~AppAPIMixIn.app_preferences`."""
        return self._client.app_preferences()

    @preferences.setter
    def preferences(self, value: Mapping[str, Any]) -> None:
        """Implements :meth:`~AppAPIMixIn.app_set_preferences`."""
        self.set_preferences(prefs=value)

    def set_preferences(
        self,
        prefs: Mapping[str, Any] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~AppAPIMixIn.app_set_preferences`."""
        self._client.app_set_preferences(prefs=prefs, **kwargs)

    setPreferences = set_preferences

    @property
    def default_save_path(self) -> str:
        """Implements :meth:`~AppAPIMixIn.app_default_save_path`."""
        return self._client.app_default_save_path()

    defaultSavePath = default_save_path

    @property
    def cookies(self) -> CookieList:
        """Implements :meth:`~AppAPIMixIn.app_cookies`."""
        return self._client.app_cookies()

    def set_cookies(
        self,
        cookies: CookieList | Sequence[Mapping[str, str | int]] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~AppAPIMixIn.app_set_cookies`."""
        self._client.app_set_cookies(cookies=cookies, **kwargs)

    setCookies = set_cookies

    @property
    def network_interface_list(self) -> NetworkInterfaceList:
        """Implements :meth:`~AppAPIMixIn.app_network_interface_list`."""
        return self._client.app_network_interface_list()

    networkInterfaceList = network_interface_list

    def network_interface_address_list(
        self,
        interface_name: str = "",
        **kwargs: APIKwargsT,
    ) -> NetworkInterfaceAddressList:
        """Implements :meth:`~AppAPIMixIn.app_network_interface_address_list`."""
        return self._client.app_network_interface_address_list(
            interface_name=interface_name,
            **kwargs,
        )

    networkInterfaceAddressList = network_interface_address_list

    def send_test_email(self) -> None:
        """Implements :meth:`~AppAPIMixIn.app_send_test_email`."""
        self._client.app_send_test_email()

    sendTestEmail = send_test_email

    def get_directory_content(
        self,
        directory_path: str | os.PathLike[AnyStr] | None = None,
        with_metadata: bool | None = None,
    ) -> DirectoryContentList:
        """Implements :meth:`~AppAPIMixIn.app_get_directory_content`."""
        return self._client.app_get_directory_content(
            directory_path=directory_path,
            with_metadata=with_metadata,
        )

    getDirectoryContent = get_directory_content



import sys

import pytest

from qbittorrentapi import APINames, NotFound404Error
from qbittorrentapi.search import (
    SearchCategoriesList,
    SearchJobDictionary,
    SearchPluginsList,
    SearchResultsDictionary,
    SearchStatusesList,
)
from tests.conftest import TORRENT2_HASH, TORRENT2_URL
from tests.utils import check, retry

PLUGIN_NAME = "therarbg"
PLUGIN_URL = "https://raw.githubusercontent.com/BurningMop/qBittorrent-Search-Plugins/refs/heads/main/therarbg.py"


@pytest.mark.skipif(sys.version_info < (3, 9), reason="removeprefix not in 3.8")
def test_methods(client):
    namespace = APINames.Search
    all_dotted_methods = set(dir(getattr(client, namespace)))

    for meth in [meth for meth in dir(client) if meth.startswith(f"{namespace}_")]:
        assert meth.removeprefix(f"{namespace}_") in all_dotted_methods


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "update_func", ["search_update_plugins", "search.update_plugins"]
)
def test_update_plugins(client, update_func):
    @retry()
    def do_test():
        client.func(update_func)()
        check(
            lambda: any(
                entry.message.startswith("Updating plugin ")
                or entry.message == "All plugins are already up to date."
                or entry.message.endswith("content was not found at the server (404)")
                for entry in reversed(client.log.main())
            ),
            True,
        )

    do_test()


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "update_func", ["search_update_plugins", "search.update_plugins"]
)
def test_update_plugins_not_implemented(client, update_func):
    with pytest.raises(NotImplementedError):
        client.func(update_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "search_func, enable_func",
    (
        ["search_plugins", "search_enable_plugin"],
        ["search.plugins", "search.enable_plugin"],
    ),
)
def test_enable_plugin(client, search_func, enable_func):
    def get_plugins():
        try:
            return client.func(search_func)()
        except TypeError:
            return client.func(search_func)

    @retry()
    def enable_plugin():
        assert isinstance(get_plugins(), SearchPluginsList)
        client.func(enable_func)(
            plugins=(p["name"] for p in get_plugins()), enable=False
        )
        check(
            lambda: (p["enabled"] for p in get_plugins()),
            True,
            reverse=True,
            negate=True,
        )
        client.func(enable_func)(
            plugins=(p["name"] for p in get_plugins()), enable=True
        )
        check(
            lambda: (p["enabled"] for p in get_plugins()),
            False,
            reverse=True,
            negate=True,
        )

    enable_plugin()


@pytest.mark.skipif_before_api_version("2.1.1")
def test_plugins_slice(client):
    assert isinstance(client.search_plugins()[1:2], SearchPluginsList)


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "enable_func", ["search_enable_plugin", "search.enable_plugin"]
)
def test_enable_plugin_not_implemented(client, enable_func):
    with pytest.raises(NotImplementedError):
        client.func(enable_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "install_func, uninstall_func",
    (
        ["search_install_plugin", "search_uninstall_plugin"],
        ["search.install_plugin", "search.uninstall_plugin"],
    ),
)
def test_install_uninstall_plugin(client, install_func, uninstall_func):
    @retry()
    def install_plugin():
        client.func(install_func)(sources=PLUGIN_URL)
        check(
            lambda: (p.name for p in client.search.plugins),
            PLUGIN_NAME,
            reverse=True,
        )

    @retry()
    def uninstall_plugin():
        client.func(uninstall_func)(names=PLUGIN_NAME)
        check(
            lambda: (p.name for p in client.search.plugins),
            PLUGIN_NAME,
            reverse=True,
            negate=True,
        )

    install_plugin()
    uninstall_plugin()


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "install_func, uninstall_func",
    (
        ["search_install_plugin", "search_uninstall_plugin"],
        ["search.install_plugin", "search.uninstall_plugin"],
    ),
)
def test_install_uninstall_plugin_not_implemented(client, install_func, uninstall_func):
    with pytest.raises(NotImplementedError):
        client.func(install_func)()
    with pytest.raises(NotImplementedError):
        client.func(uninstall_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.skipif_after_api_version("2.6")
@pytest.mark.parametrize("categories_func", ["search_categories", "search.categories"])
def test_categories(client, categories_func):
    assert isinstance(client.func(categories_func)(), SearchCategoriesList)
    assert isinstance(client.func(categories_func)()[1:2], SearchCategoriesList)
    check(lambda: client.func(categories_func)(), "All categories", reverse=True)


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize("categories_func", ["search_categories", "search.categories"])
def test_categories_not_implemented(client, categories_func):
    with pytest.raises(NotImplementedError):
        client.func(categories_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "start_func, status_func, results_func, stop_func, delete_stop",
    [
        (
            "search_start",
            "search_status",
            "search_results",
            "search_stop",
            "search_delete",
        ),
        (
            "search.start",
            "search.status",
            "search.results",
            "search.stop",
            "search.delete",
        ),
    ],
)
def test_search(client, start_func, status_func, results_func, stop_func, delete_stop):
    @retry()
    def do_test():
        job = client.func(start_func)(
            pattern="Ubuntu", plugins="enabled", category="all"
        )

        statuses = client.func(status_func)(search_id=job["id"])
        assert statuses[0]["status"] == "Running"
        assert isinstance(job, SearchJobDictionary)
        assert isinstance(statuses, SearchStatusesList)

        results = client.func(results_func)(search_id=job["id"], limit=1)
        assert isinstance(results, SearchResultsDictionary)
        results = job.results()
        assert isinstance(results, SearchResultsDictionary)

        client.func(stop_func)(search_id=job["id"])
        check(
            lambda: client.func(status_func)(search_id=job["id"])[0]["status"],
            "Stopped",
        )

        client.func(delete_stop)(search_id=job["id"])
        statuses = client.func(status_func)()
        assert not statuses

    do_test()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize("status_func", ["search_status", "search.status"])
def test_statuses_slice(client, status_func):
    assert isinstance(client.func(status_func)()[1:2], SearchStatusesList)


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func",
    [
        "search_start",
        "search_status",
        "search_results",
        "search_stop",
        "search_delete",
        "search.start",
        "search.status",
        "search.results",
        "search.stop",
        "search.delete",
    ],
)
def test_search_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        client.func(client_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "stop_func, start_func",
    [("search_stop", "search_start"), ("search.stop", "search.start")],
)
def test_stop(client, stop_func, start_func):
    job = client.func(start_func)(pattern="Ubuntu", plugins="enabled", category="all")
    check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Running")

    client.func(stop_func)(search_id=job.id)
    check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Stopped")

    job = client.func(start_func)(pattern="Ubuntu", plugins="enabled", category="all")
    check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Running")
    job.stop()
    check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Stopped")


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func", ["search_stop", "search_start", "search.stop", "search.start"]
)
def test_stop_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        client.func(client_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
def test_delete(client):
    job = client.search_start(pattern="Ubuntu", plugins="enabled", category="all")
    job.delete()
    with pytest.raises(NotFound404Error):
        job.status()


@pytest.mark.skipif_after_api_version("2.1.1")
def test_delete_not_implemented(client):
    with pytest.raises(NotImplementedError):
        client.search_stop(search_id=100)


@pytest.mark.skipif_before_api_version("2.11")
@pytest.mark.parametrize(
    "client_func",
    [
        "search_download_torrent",
        "search_downloadTorrent",
        "search.download_torrent",
        "search.downloadTorrent",
    ],
)
def test_download_torrent(client, client_func):
    # run update to ensure plugins are loaded
    client.search.update_plugins()
    check(
        lambda: [p.name for p in client.search.plugins],
        "eztv",
        reverse=True,
    )
    try:
        client.func(client_func)(url=TORRENT2_URL, plugin="eztv")
        check(
            lambda: [t.hash for t in client.torrents_info()],
            TORRENT2_HASH,
            reverse=True,
        )
    finally:
        client.torrents.delete(torrent_hashes=TORRENT2_HASH)
        check(
            lambda: [t.hash for t in client.torrents_info()],
            TORRENT2_HASH,
            reverse=True,
            negate=True,
        )


@pytest.mark.skipif_after_api_version("2.11")
def test_download_torrent_not_implemented(client):
    with pytest.raises(NotImplementedError):
        client.search_download_torrent(search_id=100)



import pytest

from qbittorrentapi import Version
from tests.conftest import IS_QBT_DEV, api_version_map


@pytest.mark.skipif(IS_QBT_DEV, reason="testing devel version of qBittorrent")
def test_supported_versions(app_version, api_version):
    assert isinstance(Version.supported_api_versions(), set)
    assert api_version in Version.supported_api_versions()
    assert isinstance(Version.supported_app_versions(), set)
    assert app_version in Version.supported_app_versions()


@pytest.mark.skipif(IS_QBT_DEV, reason="testing devel version of qBittorrent")
def test_is_supported(app_version, api_version):
    assert Version.is_app_version_supported(app_version) is True
    assert Version.is_app_version_supported(app_version[1:]) is True
    assert Version.is_app_version_supported("0.0.0") is False

    assert Version.is_api_version_supported(api_version) is True
    assert Version.is_api_version_supported("v" + api_version) is True
    assert Version.is_api_version_supported("0.0.0") is False


@pytest.mark.skipif(IS_QBT_DEV, reason="testing devel version of qBittorrent")
def test_latest_version():
    expected_latest_api_version = list(api_version_map.values())[-1]
    expected_latest_app_version = list(api_version_map.keys())[-1]
    assert Version.latest_supported_api_version() == expected_latest_api_version
    assert Version.latest_supported_app_version() == expected_latest_app_version



from __future__ import annotations

from collections.abc import Iterable

from qbittorrentapi._version_support import v
from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.definitions import (
    APIKwargsT,
    APINames,
    ClientCache,
    Dictionary,
    JsonValueT,
)


class TransferInfoDictionary(Dictionary[JsonValueT]):
    """
    Response to :meth:`~TransferAPIMixIn.transfer_info`

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-global-transfer-info>`_
    """  # noqa: E501


class TransferAPIMixIn(AppAPIMixIn):
    """
    Implementation of all ``Transfer`` API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> transfer_info = client.transfer_info()
        >>> client.transfer_set_download_limit(limit=1024000)
    """  # noqa: E501

    @property
    def transfer(self) -> Transfer:
        """
        Allows for transparent interaction with Transfer endpoints.

        See Transfer class for usage.
        """
        if self._transfer is None:
            self._transfer = Transfer(client=self)
        return self._transfer

    def transfer_info(self, **kwargs: APIKwargsT) -> TransferInfoDictionary:
        """Retrieves the global transfer info found in qBittorrent status bar."""
        return self._get_cast(
            _name=APINames.Transfer,
            _method="info",
            response_class=TransferInfoDictionary,
            **kwargs,
        )

    def transfer_speed_limits_mode(self, **kwargs: APIKwargsT) -> str:
        """Returns ``1`` if alternative speed limits are currently enabled, ``0``
        otherwise."""
        return self._get_cast(
            _name=APINames.Transfer,
            _method="speedLimitsMode",
            response_class=str,
            **kwargs,
        )

    transfer_speedLimitsMode = transfer_speed_limits_mode

    def transfer_set_speed_limits_mode(
        self,
        intended_state: bool | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Sets whether alternative speed limits are enabled.

        :param intended_state: True to enable alt speed and False to disable. Leaving
            None will toggle the current state.
        """
        if (
            intended_state is None
            or v(self.app_web_api_version()) < v("2.8.14")
            and ((self.transfer.speed_limits_mode == "1") is not bool(intended_state))
        ):
            self._post(
                _name=APINames.Transfer,
                _method="toggleSpeedLimitsMode",
                **kwargs,
            )
        else:
            data = {"mode": 1 if intended_state else 0}
            self._post(
                _name=APINames.Transfer,
                _method="setSpeedLimitsMode",
                data=data,
                **kwargs,
            )

    transfer_setSpeedLimitsMode = transfer_set_speed_limits_mode
    transfer_toggleSpeedLimitsMode = transfer_set_speed_limits_mode
    transfer_toggle_speed_limits_mode = transfer_set_speed_limits_mode

    def transfer_download_limit(self, **kwargs: APIKwargsT) -> int:
        """Retrieves download limit; 0 is unlimited."""
        return self._get_cast(
            _name=APINames.Transfer,
            _method="downloadLimit",
            response_class=int,
            **kwargs,
        )

    transfer_downloadLimit = transfer_download_limit

    def transfer_upload_limit(self, **kwargs: APIKwargsT) -> int:
        """Retrieves upload limit; 0 is unlimited."""
        return self._get_cast(
            _name=APINames.Transfer,
            _method="uploadLimit",
            response_class=int,
            **kwargs,
        )

    transfer_uploadLimit = transfer_upload_limit

    def transfer_set_download_limit(
        self,
        limit: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Set the global download limit in bytes/second.

        :param limit: download limit in bytes/second (0 or -1 for no limit)
        """
        data = {"limit": limit}
        self._post(
            _name=APINames.Transfer,
            _method="setDownloadLimit",
            data=data,
            **kwargs,
        )

    transfer_setDownloadLimit = transfer_set_download_limit

    def transfer_set_upload_limit(
        self,
        limit: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Set the global download limit in bytes/second.

        :param limit: upload limit in bytes/second (0 or -1 for no limit)
        """
        data = {"limit": limit}
        self._post(
            _name=APINames.Transfer,
            _method="setUploadLimit",
            data=data,
            **kwargs,
        )

    transfer_setUploadLimit = transfer_set_upload_limit

    def transfer_ban_peers(
        self,
        peers: str | Iterable[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Ban one or more peers.

        This method was introduced with qBittorrent v4.2.0 (Web API v2.3.0).

        :param peers: one or more peers to ban. each peer should take the form
            'host:port'
        """
        data = {"peers": self._list2string(peers, "|")}
        self._post(
            _name=APINames.Transfer,
            _method="banPeers",
            data=data,
            version_introduced="2.3",
            **kwargs,
        )

    transfer_banPeers = transfer_ban_peers


class Transfer(ClientCache[TransferAPIMixIn]):
    """
    Allows interaction with the ``Transfer`` API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> # these are all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'transfer_' prepended)
        >>> transfer_info = client.transfer.info
        >>> # access and set download/upload limits as attributes
        >>> dl_limit = client.transfer.download_limit
        >>> # this updates qBittorrent in real-time
        >>> client.transfer.download_limit = 1024000
        >>> # update speed limits mode to alternate or not
        >>> client.transfer.speedLimitsMode = True
    """  # noqa: E501

    @property
    def info(self) -> TransferInfoDictionary:
        """Implements :meth:`~TransferAPIMixIn.transfer_info`."""
        return self._client.transfer_info()

    @property
    def speed_limits_mode(self) -> str:
        """Implements :meth:`~TransferAPIMixIn.transfer_speed_limits_mode`."""
        return self._client.transfer_speed_limits_mode()

    @speed_limits_mode.setter
    def speed_limits_mode(self, val: bool) -> None:
        """Implements :meth:`~TransferAPIMixIn.transfer_set_speed_limits_mode`."""
        self.set_speed_limits_mode(intended_state=val)

    @property
    def speedLimitsMode(self) -> str:
        """Implements :meth:`~TransferAPIMixIn.transfer_speed_limits_mode`."""
        return self._client.transfer_speed_limits_mode()

    @speedLimitsMode.setter
    def speedLimitsMode(self, val: bool) -> None:
        """Implements :meth:`~TransferAPIMixIn.transfer_set_speed_limits_mode`."""
        self.set_speed_limits_mode(intended_state=val)

    def set_speed_limits_mode(
        self,
        intended_state: bool | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~TransferAPIMixIn.transfer_set_speed_limits_mode`."""
        return self._client.transfer_set_speed_limits_mode(
            intended_state=intended_state,
            **kwargs,
        )

    setSpeedLimitsMode = set_speed_limits_mode
    toggleSpeedLimitsMode = set_speed_limits_mode
    toggle_speed_limits_mode = set_speed_limits_mode

    @property
    def download_limit(self) -> int:
        """Implements :meth:`~TransferAPIMixIn.transfer_download_limit`."""
        return self._client.transfer_download_limit()

    @download_limit.setter
    def download_limit(self, val: int | str) -> None:
        """Implements :meth:`~TransferAPIMixIn.transfer_set_download_limit`."""
        self.set_download_limit(limit=val)

    @property
    def downloadLimit(self) -> int:
        """Implements :meth:`~TransferAPIMixIn.transfer_download_limit`."""
        return self._client.transfer_download_limit()

    @downloadLimit.setter
    def downloadLimit(self, val: int | str) -> None:
        """Implements :meth:`~TransferAPIMixIn.transfer_set_download_limit`."""
        self.set_download_limit(limit=val)

    @property
    def upload_limit(self) -> int:
        """Implements :meth:`~TransferAPIMixIn.transfer_upload_limit`."""
        return self._client.transfer_upload_limit()

    @upload_limit.setter
    def upload_limit(self, val: int | str) -> None:
        """Implements :meth:`~TransferAPIMixIn.transfer_set_upload_limit`."""
        self.set_upload_limit(limit=val)

    @property
    def uploadLimit(self) -> int:
        """Implements :meth:`~TransferAPIMixIn.transfer_upload_limit`."""
        return self._client.transfer_upload_limit()

    @uploadLimit.setter
    def uploadLimit(self, val: int | str) -> None:
        """Implements :meth:`~TransferAPIMixIn.transfer_set_upload_limit`."""
        self.set_upload_limit(limit=val)

    def set_download_limit(
        self,
        limit: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~TransferAPIMixIn.transfer_set_download_limit`."""
        return self._client.transfer_set_download_limit(limit=limit, **kwargs)

    setDownloadLimit = set_download_limit

    def set_upload_limit(
        self,
        limit: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~TransferAPIMixIn.transfer_set_upload_limit`."""
        return self._client.transfer_set_upload_limit(limit=limit, **kwargs)

    setUploadLimit = set_upload_limit

    def ban_peers(
        self,
        peers: str | Iterable[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """Implements :meth:`~TransferAPIMixIn.transfer_ban_peers`."""
        self._client.transfer_ban_peers(peers=peers, **kwargs)

    banPeers = ban_peers
